@@ -14,11 +14,11 @@
 #define RS_SAFEADDHASH_CMD RS_CMD_PREFIX ".SAFEADDHASH"
 #define RS_INFO_CMD RS_CMD_PREFIX ".INFO"
 #define RS_SEARCH_CMD RS_CMD_PREFIX ".SEARCH"
+#define RS_AGGREGATE_CMD RS_CMD_PREFIX ".AGGREGATE"
+
 #define RS_EXPLAIN_CMD RS_CMD_PREFIX ".EXPLAIN"
 #define RS_DEL_CMD RS_CMD_PREFIX ".DEL"
 #define RS_DROP_CMD RS_CMD_PREFIX ".DROP"
-#define RS_DTADD_CMD RS_CMD_PREFIX ".DTADD"
-#define RS_ADDTERM_CMD RS_CMD_PREFIX ".TERMADD"
 #define RS_GET_CMD RS_CMD_PREFIX ".GET"
 #define RS_MGET_CMD RS_CMD_PREFIX ".MGET"
 #define RS_TAGVALS_CMD RS_CMD_PREFIX ".TAGVALS"
@@ -28,4 +28,15 @@
 #define RS_SUGDEL_CMD RS_CMD_PREFIX ".SUGDEL"
 #define RS_SUGLEN_CMD RS_CMD_PREFIX ".SUGLEN"
 
-#endif
\ No newline at end of file
+#define RS_CURSOR_CMD RS_CMD_PREFIX ".CURSOR"
+
+#define RS_SYNADD_CMD RS_CMD_PREFIX ".SYNADD"
+#define RS_SYNUPDATE_CMD RS_CMD_PREFIX ".SYNUPDATE"
+#define RS_SYNFORCEUPDATE_CMD RS_CMD_PREFIX ".SYNFORCEUPDATE"
+#define RS_SYNDUMP_CMD RS_CMD_PREFIX ".SYNDUMP"
+
+#define RS_ALTER_CMD RS_CMD_PREFIX ".ALTER"
+
+#define RS_DEBUG RS_CMD_PREFIX ".DEBUG"
+
+#endif
