@@ -15,6 +15,7 @@
 #include "rmalloc.h"
 #include "indexer.h"
 #include "tag_index.h"
+#include "aggregate/expr/expression.h"
 
 // Memory pool for RSAddDocumentContext contexts
 static mempool_t *actxPool_g = NULL;
@@ -54,11 +55,13 @@ static int AddDocumentCtx_SetDocument(RSAddDocumentCtx *aCtx, IndexSpec *sp, Doc
 
   // size: uint16_t * SPEC_MAX_FIELDS
   FieldSpecDedupeArray dedupe = {0};
+  int hasTextFields = 0;
+  int hasOtherFields = 0;
 
   for (int i = 0; i < doc->numFields; i++) {
     const DocumentField *f = doc->fields + i;
     FieldSpec *fs = IndexSpec_GetField(sp, f->name, strlen(f->name));
-    if (fs) {
+    if (fs && f->text) {
 
       aCtx->fspecs[i] = *fs;
       if (dedupe[fs->index]) {
@@ -72,17 +75,14 @@ static int AddDocumentCtx_SetDocument(RSAddDocumentCtx *aCtx, IndexSpec *sp, Doc
         // mark sortable fields to be updated in the state flags
         aCtx->stateFlags |= ACTX_F_SORTABLES;
       }
-      if (fs->type == FIELD_FULLTEXT && FieldSpec_IsIndexable(fs)) {
-        numIndexable++;
-      }
-      // mark non text fields in the state flags
-      if (fs->type != FIELD_FULLTEXT) {
-        aCtx->stateFlags |= ACTX_F_NONTXTFLDS;
-      }
 
-      // mark indexable fields in the state flags
       if (FieldSpec_IsIndexable(fs)) {
-        aCtx->stateFlags |= ACTX_F_INDEXABLES;
+        if (fs->type == FIELD_FULLTEXT) {
+          numIndexable++;
+          hasTextFields = 1;
+        } else {
+          hasOtherFields = 1;
+        }
       }
     } else {
       // Field is not in schema, or is duplicate
@@ -90,6 +90,16 @@ static int AddDocumentCtx_SetDocument(RSAddDocumentCtx *aCtx, IndexSpec *sp, Doc
     }
   }
 
+  if (hasTextFields || hasOtherFields) {
+    aCtx->stateFlags |= ACTX_F_INDEXABLES;
+  }
+  if (!hasTextFields) {
+    aCtx->stateFlags |= ACTX_F_TEXTINDEXED;
+  }
+  if (!hasOtherFields) {
+    aCtx->stateFlags |= ACTX_F_OTHERINDEXED;
+  }
+
   if ((aCtx->stateFlags & ACTX_F_SORTABLES) && aCtx->sv == NULL) {
     aCtx->sv = NewSortingVector(sp->sortables->len);
   }
@@ -135,6 +145,14 @@ RSAddDocumentCtx *NewAddDocumentCtx(IndexSpec *sp, Document *b, const char **err
     aCtx->fwIdx = NewForwardIndex(&aCtx->doc, sp->flags);
   }
 
+  if (sp->smap) {
+    // we get a read only copy of the synonym map for accessing in the index thread with out worring
+    // about thready safe issues
+    aCtx->fwIdx->smap = SynonymMap_GetReadOnlyCopy(sp->smap);
+  } else {
+    aCtx->fwIdx->smap = NULL;
+  }
+
   aCtx->tokenizer = GetTokenizer(b->language, aCtx->fwIdx->stemmer, sp->stopwords);
   StopWordList_Ref(sp->stopwords);
 
@@ -172,35 +190,67 @@ void AddDocumentCtx_Finish(RSAddDocumentCtx *aCtx) {
 // How many bytes in a document to warrant it being tokenized in a separate thread
 #define SELF_EXEC_THRESHOLD 1024
 
+void Document_Dump(const Document *doc) {
+  printf("Document Key: %s. ID=%llu\n", RedisModule_StringPtrLen(doc->docKey, NULL), doc->docId);
+  for (size_t ii = 0; ii < doc->numFields; ++ii) {
+    printf("  [%lu]: %s => %s\n", ii, doc->fields[ii].name,
+           RedisModule_StringPtrLen(doc->fields[ii].text, NULL));
+  }
+}
+
 static void AddDocumentCtx_UpdateNoIndex(RSAddDocumentCtx *aCtx, RedisSearchCtx *sctx);
+
+static int AddDocumentCtx_ReplaceMerge(RSAddDocumentCtx *aCtx, RedisSearchCtx *sctx) {
+  /**
+   * The REPLACE operation contains fields which must be reindexed. This means
+   * that a new document ID needs to be assigned, and as a consequence, all
+   * fields must be reindexed.
+   */
+  // Free the old field data
+  size_t oldFieldCount = aCtx->doc.numFields;
+  Document_ClearDetachedFields(&aCtx->doc, sctx->redisCtx);
+
+  // Get a list of fields needed to be loaded for reindexing
+  const char **toLoad = array_new(const char *, 8);
+
+  for (size_t ii = 0; ii < sctx->spec->numFields; ++ii) {
+    // TODO: We should only need to reload fields that are actually being reindexed!
+    toLoad = array_append(toLoad, sctx->spec->fields[ii].name);
+  }
+
+  // for (size_t ii = 0; ii < array_len(toLoad); ++ii) {
+  //   printf("Loading: %s\n", toLoad[ii]);
+  // }
+
+  int rv =
+      Redis_LoadDocumentEx(sctx, aCtx->doc.docKey, toLoad, array_len(toLoad), &aCtx->doc, NULL);
+  // int rv = Redis_LoadDocument(sctx, aCtx->doc.docKey, &aCtx->doc);
+  array_free(toLoad);
+
+  if (rv != REDISMODULE_OK) {
+    RedisModule_ReplyWithError(sctx->redisCtx, "Error updating document");
+    AddDocumentCtx_Free(aCtx);
+    return 1;
+  }
+
+  // Keep hold of the new fields.
+  Document_DetachFields(&aCtx->doc, sctx->redisCtx);
+  AddDocumentCtx_SetDocument(aCtx, sctx->spec, &aCtx->doc, oldFieldCount);
+
+  return 0;
+}
+
 static int handlePartialUpdate(RSAddDocumentCtx *aCtx, RedisSearchCtx *sctx) {
   // Handle partial update of fields
   if (aCtx->stateFlags & ACTX_F_INDEXABLES) {
-
-    // The document update is partial - but it contains indexable fields. We need to load the
-    // current version of it, and then do a normal indexing flow
-
-    // Free the old field data
-    size_t oldFieldCount = aCtx->doc.numFields;
-    Document_ClearDetachedFields(&aCtx->doc, sctx->redisCtx);
-    if (Redis_LoadDocument(sctx, aCtx->doc.docKey, &aCtx->doc) != REDISMODULE_OK) {
-      RedisModule_ReplyWithError(sctx->redisCtx, "Error updating document");
-      AddDocumentCtx_Free(aCtx);
-      return 1;
-    }
-
-    // Keep hold of the new fields.
-    Document_DetachFields(&aCtx->doc, sctx->redisCtx);
-    AddDocumentCtx_SetDocument(aCtx, sctx->spec, &aCtx->doc, oldFieldCount);
-
-    return 0;
+    return AddDocumentCtx_ReplaceMerge(aCtx, sctx);
+  } else {
+    // No indexable fields are updated, we can just update the metadata.
+    // Quick update just updates the score, payload and sortable fields of the document.
+    // Thus full-reindexing of the document is not required
+    AddDocumentCtx_UpdateNoIndex(aCtx, sctx);
+    return 1;
   }
-
-  // No indexable fields are updated, we can just update the metadata.
-  // Quick update just updates the score, payload and sortable fields of the document.
-  // Thus full-reindexing of the document is not required
-  AddDocumentCtx_UpdateNoIndex(aCtx, sctx);
-  return 1;
 }
 
 void AddDocumentCtx_Submit(RSAddDocumentCtx *aCtx, RedisSearchCtx *sctx, uint32_t options) {
@@ -251,6 +301,11 @@ void AddDocumentCtx_Free(RSAddDocumentCtx *aCtx) {
     aCtx->tokenizer = NULL;
   }
 
+  if (aCtx->oldMd) {
+    DMD_Decref(aCtx->oldMd);
+    aCtx->oldMd = NULL;
+  }
+
   ByteOffsetWriter_Cleanup(&aCtx->offsetsWriter);
 
   mempool_release(actxPool_g, aCtx);
@@ -260,9 +315,14 @@ void AddDocumentCtx_Free(RSAddDocumentCtx *aCtx) {
   static int name(RSAddDocumentCtx *aCtx, const DocumentField *field, const FieldSpec *fs, \
                   fieldData *fdata, const char **errorString)
 
-#define FIELD_INDEXER(name)                                                                \
-  static int name(RSAddDocumentCtx *aCtx, RedisSearchCtx *ctx, const DocumentField *field, \
-                  const FieldSpec *fs, fieldData *fdata, const char **errorString)
+#define FIELD_BULK_INDEXER(name)                                                    \
+  static int name(IndexBulkData *bulk, RSAddDocumentCtx *aCtx, RedisSearchCtx *ctx, \
+                  DocumentField *field, const FieldSpec *fs, fieldData *fdata,      \
+                  const char **errorString)
+
+#define FIELD_BULK_CTOR(name) \
+  static void name(IndexBulkData *bulk, const FieldSpec *fs, RedisSearchCtx *ctx)
+#define FIELD_BULK_FINALIZER(name) static void name(IndexBulkData *bulk, RedisSearchCtx *ctx)
 
 #define FIELD_PREPROCESSOR FIELD_HANDLER
 
@@ -317,14 +377,19 @@ FIELD_PREPROCESSOR(numericPreprocessor) {
   return 0;
 }
 
-FIELD_INDEXER(numericIndexer) {
-  RedisModuleKey *idxKey;
-  NumericRangeTree *rt = OpenNumericIndex(ctx, fs->name, &idxKey);
+FIELD_BULK_INDEXER(numericIndexer) {
+  NumericRangeTree *rt = bulk->indexData;
+  // NumericRangeTree *rt = OpenNumericIndex(ctx, fs->name, &idxKey);
   NumericRangeTree_Add(rt, aCtx->doc.docId, fdata->numeric);
-  RedisModule_CloseKey(idxKey);
+  // RedisModule_CloseKey(idxKey);
   return 0;
 }
 
+FIELD_BULK_CTOR(numericCtor) {
+  RedisModuleString *keyName = IndexSpec_GetFormattedKey(ctx->spec, fs);
+  bulk->indexData = OpenNumericIndex(ctx, keyName, &bulk->indexKey);
+}
+
 FIELD_PREPROCESSOR(geoPreprocessor) {
   const char *c = RedisModule_StringPtrLen(field->text, NULL);
   char *pos = strpbrk(c, " ,");
@@ -339,7 +404,7 @@ FIELD_PREPROCESSOR(geoPreprocessor) {
   return 0;
 }
 
-FIELD_INDEXER(geoIndexer) {
+FIELD_BULK_INDEXER(geoIndexer) {
   GeoIndex gi = {.ctx = ctx, .sp = fs};
   int rv = GeoIndex_AddStrings(&gi, aCtx->doc.docId, fdata->geo.slon, fdata->geo.slat);
 
@@ -353,35 +418,33 @@ FIELD_INDEXER(geoIndexer) {
 FIELD_PREPROCESSOR(tagPreprocessor) {
 
   fdata->tags = TagIndex_Preprocess(&fs->tagOpts, field);
+
   if (fdata->tags == NULL) {
-    *errorString = "Could not index tag field";
-    return -1;
+    return 0;
+  }
+  if (FieldSpec_IsSortable(fs)) {
+    size_t fl;
+    const char *c = RedisModule_StringPtrLen(field->text, &fl);
+    RSSortingVector_Put(aCtx->sv, fs->sortIdx, (void *)c, RS_SORTABLE_STR);
   }
   return 0;
 }
 
-FIELD_INDEXER(tagIndexer) {
-  RedisModuleKey *idxKey;
-  RedisModuleString *kname = TagIndex_FormatName(ctx, fs->name);
+FIELD_BULK_CTOR(tagCtor) {
+  RedisModuleString *kname = IndexSpec_GetFormattedKey(ctx->spec, fs);
+  bulk->indexData = TagIndex_Open(ctx->redisCtx, kname, 1, &bulk->indexKey);
+}
+
+FIELD_BULK_INDEXER(tagIndexer) {
   int rc = 0;
-  TagIndex *ti = TagIndex_Open(ctx->redisCtx, kname, 1, &idxKey);
-  if (!ti) {
+  if (!bulk->indexData) {
     *errorString = "Could not open tag index for indexing";
     rc = -1;
-    goto cleanup;
+  } else {
+    TagIndex_Index(bulk->indexData, fdata->tags, aCtx->doc.docId);
   }
-
-  TagIndex_Index(ti, fdata->tags, aCtx->doc.docId);
-cleanup:
-  RedisModule_CloseKey(idxKey);
-  RedisModule_FreeString(ctx->redisCtx, kname);
   if (fdata->tags) {
-    for (size_t i = 0; i < Vector_Size(fdata->tags); i++) {
-      char *tok = NULL;
-      Vector_Get(fdata->tags, i, &tok);
-      free(tok);
-    }
-    Vector_Free(fdata->tags);
+    TagIndex_FreePreprocessedData(fdata->tags);
   }
   return rc;
 }
@@ -401,16 +464,20 @@ PreprocessorFunc GetIndexPreprocessor(const FieldType ft) {
   }
 }
 
-IndexerFunc GetIndexIndexer(const FieldType ft) {
+static BulkIndexer geoBulkProcs = {.BulkAdd = geoIndexer};
+static BulkIndexer numBulkProcs = {.BulkInit = numericCtor, .BulkAdd = numericIndexer};
+static BulkIndexer tagBulkProcs = {.BulkInit = tagCtor, .BulkAdd = tagIndexer};
+
+const BulkIndexer *GetBulkIndexer(const FieldType ft) {
   switch (ft) {
     case FIELD_NUMERIC:
-      return numericIndexer;
-    case FIELD_GEO:
-      return geoIndexer;
+      return &numBulkProcs;
     case FIELD_TAG:
-      return tagIndexer;
-    case FIELD_FULLTEXT:
+      return &tagBulkProcs;
+    case FIELD_GEO:
+      return &geoBulkProcs;
     default:
+      abort();
       return NULL;
   }
 }
@@ -454,6 +521,93 @@ cleanup:
   return ourRv;
 }
 
+/* Evaluate an IF expression (e.g. IF "@foo == 'bar'") against a document, by getting the properties
+ * from the sorting table or from the hash representation of the document.
+ *
+ * NOTE: This is disconnected from the document indexing flow, and loads the document and discards
+ * of it internally
+ *
+ * Returns  REDISMODULE_ERR on failure, OK otherwise*/
+int Document_EvalExpression(RedisSearchCtx *sctx, RedisModuleString *key, const char *expr,
+                            int *result, char **err) {
+
+  // Try to parser the expression first, fail if we can't
+  RSExpr *e = RSExpr_Parse(expr, strlen(expr), err);
+  if (!e) {
+    return REDISMODULE_ERR;
+  }
+
+  // Get the fields needed to evaluate the expression, so we'll know what to load (if any)
+  const char **fields = Expr_GetRequiredFields(e);
+
+  Document doc = {.docKey = key};
+  // Get the metadata, which should include sortables
+  RSDocumentMetadata *md = DocTable_GetByKeyR(&sctx->spec->docs, doc.docKey);
+
+  // Make sure the field list only includes fields which are not already in the sorting vector
+  size_t loadFields = 0;
+  if (md && md->sortVector) {
+    // If a field is in the sorting vector, we simply skip it. n is the total fields not in the
+    // sorting vector
+    for (size_t i = 0; i < array_len(fields); i++) {
+      if (RSSortingTable_GetFieldIdx(sctx->spec->sortables, fields[i]) == -1) {
+        fields[loadFields++] = fields[i];
+      }
+    }
+  } else {
+    // If we don't have a sorting vector - we need to load all the fields.
+    loadFields = array_len(fields);
+  }
+
+  // loadFields > 0 means that some fields needed are not sortable and should be loaded from hash
+  if (loadFields > 0) {
+    if (Redis_LoadDocumentEx(sctx, key, fields, loadFields, &doc, NULL) == REDISMODULE_ERR) {
+      SET_ERR(err, "Could not load document");
+      array_free(fields);
+      return REDISMODULE_ERR;
+    }
+  }
+
+  // Create a field map from the document fields
+  RSFieldMap *fm = RS_NewFieldMap(doc.numFields);
+  for (int i = 0; i < doc.numFields; i++) {
+    RSFieldMap_Add(&fm, doc.fields[i].name, RS_RedisStringVal(doc.fields[i].text));
+  }
+
+  // create a mock search result
+  SearchResult res = (SearchResult){
+      .docId = doc.docId,
+      .fields = fm,
+      .scorerPrivateData = md,
+  };
+  // All this is needed to eval the expression
+  RSFunctionEvalCtx *fctx = RS_NewFunctionEvalCtx();
+  fctx->res = &res;
+  RSExprEvalCtx evctx = (RSExprEvalCtx){
+      .r = &res,
+      .sortables = sctx ? (sctx->spec ? sctx->spec->sortables : NULL) : NULL,
+      .fctx = fctx,
+  };
+  RSValue out = RSVALUE_STATIC;
+  int rc = REDISMODULE_OK;
+
+  // Now actually eval the expression
+  if (EXPR_EVAL_ERR == RSExpr_Eval(&evctx, e, &out, err)) {
+    rc = REDISMODULE_ERR;
+  } else {
+    // The result is the boolean value of the expression's output
+    *result = RSValue_BoolTest(&out);
+  }
+
+  // Cleanup
+  array_free(fields);
+  RSFunctionEvalCtx_Free(fctx);
+  RSFieldMap_Free(fm);
+  RSExpr_Free(e);
+  Document_Free(&doc);
+  return rc;
+}
+
 static void AddDocumentCtx_UpdateNoIndex(RSAddDocumentCtx *aCtx, RedisSearchCtx *sctx) {
 #define BAIL(s)            \
   do {                     \
@@ -530,13 +684,6 @@ done:
   AddDocumentCtx_Free(aCtx);
 }
 
-int Document_CanAdd(Document *doc, IndexSpec *sp, int replace) {
-  if (!replace && DocTable_GetId(&sp->docs, MakeDocKeyR(doc->docKey)) != 0) {
-    return 0;
-  }
-  return 1;
-}
-
 DocumentField *Document_GetField(Document *d, const char *fieldName) {
   if (!d || !fieldName) return NULL;
 
@@ -546,4 +693,4 @@ DocumentField *Document_GetField(Document *d, const char *fieldName) {
     }
   }
   return NULL;
-}
\ No newline at end of file
+}
