@@ -7,7 +7,6 @@
 
 #include "commands.h"
 #include "document.h"
-#include "version.h"
 #include "forward_index.h"
 #include "geo_index.h"
 #include "tag_index.h"
@@ -30,7 +29,11 @@
 #include "search_request.h"
 #include "config.h"
 #include "gc.h"
+#include "aggregate/aggregate.h"
 #include "rmalloc.h"
+#include "cursor.h"
+#include "version.h"
+#include "debug_commads.h"
 
 #define LOAD_INDEX(ctx, srcname, write)                                                     \
   ({                                                                                        \
@@ -59,10 +62,8 @@ static int CheckConcurrentSupport(RedisModuleCtx *ctx) {
 }
 
 /*
-## FT.ADD <index> <docId> <score> [NOSAVE] [REPLACE] [LANGUAGE <lang>] [PAYLOAD {payload}] FIELDS
-<field>
-<text> ....]
-Add a documet to the index.
+## FT.ADD <index> <docId> <score> [NOSAVE] [REPLACE] [PARTIAL] [IF <expr>] [LANGUAGE <lang>]
+[PAYLOAD {payload}] FIELDS <field> <text> ....] Add a documet to the index.
 
 ## Parameters:
 
@@ -103,7 +104,8 @@ English.
    > "russian", "spanish",   "swedish", "tamil",     "turkish"
 
 
-Returns OK on success, or an error if something went wrong.
+Returns OK on success, NOADD if the document was not added due to an IF expression not evaluating to
+true or an error if something went wrong.
 */
 static int doAddDocument(RedisModuleCtx *ctx, RedisModuleString **argv, int argc, int canBlock) {
   int nosave = RMUtil_ArgExists("NOSAVE", argv, argc, 1);
@@ -158,14 +160,37 @@ static int doAddDocument(RedisModuleCtx *ctx, RedisModuleString **argv, int argc
     RedisModule_ReplyWithError(ctx, "Unknown index name");
     goto cleanup;
   }
+  RedisSearchCtx sctx = {.redisCtx = ctx, .spec = sp};
 
-  Document doc;
-  Document_PrepareForAdd(&doc, argv[2], ds, argv, fieldsIdx, argc, lang, payload, ctx);
-  if (!Document_CanAdd(&doc, sp, replace)) {
-    Document_FreeDetached(&doc, ctx);
+  // If the ID is 0, then the document does not exist.
+  int exists = !!DocTable_GetId(&sp->docs, MakeDocKeyR(argv[2]));
+  if (exists && !replace) {
     RedisModule_ReplyWithError(ctx, "Document already in index");
     goto cleanup;
   }
+
+  // Parse optional IF "expr" flag before FIELDS, only if the document exists
+  const char *expr = NULL;
+  if (exists && RMUtil_ParseArgsAfter("IF", argv, fieldsIdx, "c", &expr) == REDISMODULE_OK) {
+    char *err = NULL;
+    int res = 0;
+    if (Document_EvalExpression(&sctx, argv[2], expr, &res, &err) == REDISMODULE_OK) {
+      if (res == 0) {
+        RedisModule_ReplyWithSimpleString(ctx, "NOADD");
+        goto cleanup;
+      }
+    } else {
+      char buf[1024];
+      snprintf(buf, sizeof(buf), "Could not evaluate IF expression: %s", err);
+      RedisModule_ReplyWithError(ctx, buf);
+      ERR_FREE(err);
+      goto cleanup;
+    }
+  }
+
+  Document doc;
+  Document_PrepareForAdd(&doc, argv[2], ds, argv, fieldsIdx, argc, lang, payload, ctx);
+
   if (!nosave) {
     RedisSearchCtx sctx = SEARCH_CTX_STATIC(ctx, sp);
     if (Redis_SaveDocument(&sctx, &doc) != REDISMODULE_OK) {
@@ -185,16 +210,18 @@ static int doAddDocument(RedisModuleCtx *ctx, RedisModuleString **argv, int argc
 
   // in partial mode
   uint32_t options = 0;
-  if (replace) {
-    options |= DOCUMENT_ADD_REPLACE;
-  }
-  if (partial) {
-    options |= DOCUMENT_ADD_PARTIAL;
+  if (exists) {
+    if (replace) {
+      options |= DOCUMENT_ADD_REPLACE;
+    }
+    if (partial) {
+      options |= DOCUMENT_ADD_PARTIAL;
+    }
   }
   if (nosave) {
     options |= DOCUMENT_ADD_NOSAVE;
   }
-  RedisSearchCtx sctx = {.redisCtx = ctx, .spec = sp};
+
   if (!canBlock) {
     aCtx->stateFlags |= ACTX_F_NOBLOCK;
   }
@@ -270,9 +297,13 @@ static int renderIndexOptions(RedisModuleCtx *ctx, IndexSpec *sp) {
   RedisModule_ReplyWithSimpleString(ctx, "index_options");
   RedisModule_ReplyWithArray(ctx, REDISMODULE_POSTPONED_ARRAY_LEN);
   int n = 0;
-  ADD_NEGATIVE_OPTION(Index_StoreFreqs, "NOFREQS");
-  ADD_NEGATIVE_OPTION(Index_StoreFieldFlags, "NOFIELDS");
-  ADD_NEGATIVE_OPTION(Index_StoreTermOffsets, "NOOFFSETS");
+  ADD_NEGATIVE_OPTION(Index_StoreFreqs, SPEC_NOFREQS_STR);
+  ADD_NEGATIVE_OPTION(Index_StoreFieldFlags, SPEC_NOFIELDS_STR);
+  ADD_NEGATIVE_OPTION(Index_StoreTermOffsets, SPEC_NOOFFSETS_STR);
+  if (sp->flags & Index_WideSchema) {
+    RedisModule_ReplyWithSimpleString(ctx, SPEC_SCHEMA_EXPANDABLE_STR);
+    n++;
+  }
   RedisModule_ReplySetArrayLength(ctx, n);
   return 2;
 }
@@ -343,6 +374,8 @@ int IndexInfoCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
   //  REPLY_KVNUM(n, "score_index_size_mb", sp->stats.scoreIndexesSize / (float)0x100000);
 
   REPLY_KVNUM(n, "doc_table_size_mb", sp->docs.memsize / (float)0x100000);
+  REPLY_KVNUM(n, "sortable_values_size_mb", sp->docs.sortablesSize / (float)0x100000);
+
   REPLY_KVNUM(n, "key_table_size_mb", TrieMap_MemUsage(sp->docs.dim.tm) / (float)0x100000);
   REPLY_KVNUM(n, "records_per_doc_avg",
               (float)sp->stats.numRecords / (float)sp->stats.numDocuments);
@@ -357,6 +390,10 @@ int IndexInfoCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
   GC_RenderStats(ctx, sp->gc);
   n += 2;
 
+  RedisModule_ReplyWithSimpleString(ctx, "cursor_stats");
+  Cursors_RenderStats(&RSCursors, sp->name, ctx);
+  n += 2;
+
   RedisModule_ReplySetArrayLength(ctx, n);
   return REDISMODULE_OK;
 }
@@ -441,29 +478,29 @@ int QueryExplainCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc)
     return RedisModule_ReplyWithError(ctx, "Unknown Index name");
   }
 
-  char *err;
+  char *err = NULL;
 
   RSSearchRequest *req = ParseRequest(sctx, argv, argc, &err);
   if (req == NULL) {
     RedisModule_Log(ctx, "warning", "Error parsing request: %s", err);
     SearchCtx_Free(sctx);
-    return RedisModule_ReplyWithError(ctx, err);
+    RedisModule_ReplyWithError(ctx, err);
+    ERR_FREE(err);
+    return REDISMODULE_OK;
   }
-  req->sctx = sctx;
 
-  QueryParseCtx *q = NewQueryParseCtx(req);
+  QueryParseCtx *q = NewQueryParseCtx(sctx, req->rawQuery, req->qlen, &req->opts);
   if (!q) {
     SearchCtx_Free(sctx);
     return RedisModule_ReplyWithError(ctx, "Error parsing query");
   }
 
-  char *errMsg = NULL;
-  if (!Query_Parse(q, &errMsg)) {
+  if (!Query_Parse(q, &err)) {
 
-    if (errMsg) {
-      RedisModule_Log(ctx, "debug", "Error parsing query: %s", errMsg);
-      RedisModule_ReplyWithError(ctx, errMsg);
-      free(errMsg);
+    if (err) {
+      RedisModule_Log(ctx, "debug", "Error parsing query: %s", err);
+      RedisModule_ReplyWithError(ctx, err);
+      ERR_FREE(err);
     } else {
       /* Simulate an empty response - this means an empty query */
       RedisModule_ReplyWithArray(ctx, 1);
@@ -471,8 +508,9 @@ int QueryExplainCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc)
     }
     goto end;
   }
-  if (!(req->flags & Search_Verbatim)) {
-    Query_Expand(q, req->expander);
+
+  if (!(req->opts.flags & Search_Verbatim)) {
+    Query_Expand(q, req->opts.expander);
   }
   if (req->geoFilter) {
     Query_SetGeoFilter(q, req->geoFilter);
@@ -506,103 +544,26 @@ end:
   return REDISMODULE_OK;
 }
 
-/* FT.DTADD {index} {key} {flags} {score} {payload} {byteOffsets}
- *
- *  **WARNING**:  Do NOT use this command, it is for internal use in AOF rewriting only!!!!
- *
- *  This command is used only for AOF rewrite and makes sure the document table is rebuilt in the
- *  same order as as in memory
- *
- *  Returns the docId on success or 0 on failure
- */
-int DTAddCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
-  RedisModule_AutoMemory(ctx);
-  if (argc < 7) {
-    return RedisModule_WrongArity(ctx);
+#define GEN_CONCURRENT_WRAPPER(name, argcond, target, pooltype)                      \
+  static int name(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {         \
+    if (!(argcond)) {                                                                \
+      return RedisModule_WrongArity(ctx);                                            \
+    }                                                                                \
+    if (CheckConcurrentSupport(ctx)) {                                               \
+      return ConcurrentSearch_HandleRedisCommand(pooltype, target, ctx, argv, argc); \
+    } else {                                                                         \
+      target(ctx, argv, argc, NULL);                                                 \
+      return REDISMODULE_OK;                                                         \
+    }                                                                                \
   }
 
-  IndexSpec *sp = IndexSpec_Load(ctx, RedisModule_StringPtrLen(argv[1], NULL), 1);
-  if (sp == NULL) {
-    return RedisModule_ReplyWithError(ctx, "Unknown Index name");
-  }
+GEN_CONCURRENT_WRAPPER(CursorCommand, argc >= 4, AggregateCommand_ExecCursor,
+                       CONCURRENT_POOL_SEARCH)
 
-  long long flags;
-  double score;
-  if (RMUtil_ParseArgs(argv, argc, 3, "ld", &flags, &score) == REDISMODULE_ERR) {
-    return RedisModule_ReplyWithError(ctx, "Could not parse flags and score");
-  }
+GEN_CONCURRENT_WRAPPER(AggregateCommand, argc >= 3, AggregateCommand_ExecAggregate,
+                       CONCURRENT_POOL_SEARCH);
 
-  size_t payloadSize = 0, offsetsSize = 0, svSize = 0;
-
-  const char *payload = RedisModule_StringPtrLen(argv[5], &payloadSize);
-  const char *serOffsets = RedisModule_StringPtrLen(argv[6], &offsetsSize);
-  const char *svBuf = argc >= 8 ? RedisModule_StringPtrLen(argv[7], &svSize) : NULL;
-  RSSortingVector *sv = NULL;
-  RSByteOffsets *offsets = NULL;
-
-  if (flags & Document_HasSortVector) {
-    // Check if we're actually passed an SV
-    if (svBuf == NULL || (sv = SortingVector_LoadSerialized(svBuf, svSize)) == NULL) {
-      flags &= ~Document_HasSortVector;
-    }
-  }
-  if (flags & Document_HasOffsetVector) {
-    if (offsetsSize) {
-      Buffer *b = Buffer_Wrap((char *)serOffsets, offsetsSize);
-      offsets = LoadByteOffsets(b);
-      free(b);
-    }
-    if (!offsets) {
-      flags &= ~Document_HasOffsetVector;
-    }
-  }
-
-  t_docId d = DocTable_Put(&sp->docs, MakeDocKeyR(argv[2]), (float)score, (u_char)flags, payload,
-                           payloadSize);
-
-  if (offsets) {
-    DocTable_SetByteOffsets(&sp->docs, d, offsets);
-  }
-
-  if (sv) {
-    DocTable_SetSortingVector(&sp->docs, d, sv);
-  }
-
-  return RedisModule_ReplyWithLongLong(ctx, d);
-}
-
-/**
- * FT.TERMADD {index} {term} {score}
- *
- * **WARNING** Do NOT use this command. It is for internal AOF rewriting only
- *
- * This command is used to incrementally transfer terms (for prefix expansion)
- * over to the trie.
- *
- * This might change once the internal structure of the trie changes
- */
-int TermAddCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
-  RedisModule_AutoMemory(ctx);
-  if (argc != 4) {
-    return RedisModule_WrongArity(ctx);
-  }
-
-  IndexSpec *sp = LOAD_INDEX(ctx, argv[1], 1);
-
-  // Add the term to the spec
-  size_t termLen = 0;
-  const char *termStr = RedisModule_StringPtrLen(argv[2], &termLen);
-
-  double score;
-  if (RedisModule_StringToDouble(argv[3], &score) != REDISMODULE_OK) {
-    return RedisModule_ReplyWithError(ctx, "ERR bad score");
-  }
-
-  IndexSpec_RestoreTerm(sp, termStr, termLen, score);
-  return REDISMODULE_OK;
-}
-
-/* FT.DEL {index} {doc_id} [DD]
+/* FT.DEL {index} {doc_id}
  *  Delete a document from the index. Returns 1 if the document was in the index, or 0 if not.
  *
  *  **NOTE**: This does not actually delete the document from the index, just marks it as deleted
@@ -833,37 +794,67 @@ and
 then pairs of
     document id, and a nested array of field/value, unless NOCONTENT was given
 */
-int SearchCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
+void _SearchCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc,
+                    struct ConcurrentCmdCtx *cmdCtx) {
   // at least one field, and number of field/text args must be even
-  if (argc < 3) {
-    return RedisModule_WrongArity(ctx);
-  }
 
   RedisModule_AutoMemory(ctx);
   RedisSearchCtx *sctx = NewSearchCtx(ctx, argv[1]);
   if (sctx == NULL) {
-    return RedisModule_ReplyWithError(ctx, "Unknown Index name");
+    RedisModule_ReplyWithError(ctx, "Unknown Index name");
+    return;
   }
 
-  char *err;
+  char *err = NULL;
+  RSSearchRequest *req = NULL;
+  QueryParseCtx *q = NULL;
+  QueryPlan *plan = NULL;
 
-  RSSearchRequest *req = ParseRequest(sctx, argv, argc, &err);
+  req = ParseRequest(sctx, argv, argc, &err);
   if (req == NULL) {
     RedisModule_Log(ctx, "warning", "Error parsing request: %s", err);
-    SearchCtx_Free(sctx);
-    return RedisModule_ReplyWithError(ctx, err);
+    RedisModule_ReplyWithError(ctx, err ? err : "Error parsing request");
+
+    goto end;
   }
 
-  // in concurrent mode - process the request in the thread pool
-  if (CheckConcurrentSupport(ctx)) {
-    int rc = RSSearchRequest_ProcessInThreadpool(ctx, req);
-    SearchCtx_Free(sctx);
-    return rc;
-  } else {  // "safe" mode - process the request in the main thread
-    return RSSearchRequest_ProcessMainThread(sctx, req);
+  q = SearchRequest_ParseQuery(sctx, req, &err);
+  if (!q && err) {
+    RedisModule_Log(ctx, "warning", "Error parsing query: %s", err);
+    RedisModule_ReplyWithError(ctx, err);
+    goto end;
   }
+
+  plan = SearchRequest_BuildPlan(sctx, req, q, &err);
+  if (!plan) {
+    if (err) {
+      RedisModule_Log(ctx, "debug", "Error parsing query: %s", err);
+      RedisModule_ReplyWithError(ctx, err);
+      free(err);
+    } else {
+      /* Simulate an empty response - this means an empty query */
+      RedisModule_ReplyWithArray(ctx, 1);
+      RedisModule_ReplyWithLongLong(ctx, 0);
+    }
+    goto end;
+  }
+
+  QueryPlan_Run(plan, ctx);
+  if (err) {
+    RedisModule_ReplyWithError(ctx, err);
+  }
+
+end:
+  ERR_FREE(err);
+
+  if (plan) QueryPlan_Free(plan);
+  if (sctx) SearchCtx_Free(sctx);
+  if (req) RSSearchRequest_Free(req);
+  if (q) Query_Free(q);
 }
 
+GEN_CONCURRENT_WRAPPER(SearchCommand, argc >= 3, _SearchCommand, CONCURRENT_POOL_SEARCH)
+
 /* FT.TAGVALS {idx} {field}
  * Return all the values of a tag field.
  * There is no sorting or paging, so be careful with high-cradinality tag fields */
@@ -947,7 +938,10 @@ int CreateIndexCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc)
 
   IndexSpec *sp = IndexSpec_CreateNew(ctx, argv, argc, &err);
   if (sp == NULL) {
-    return RedisModule_ReplyWithError(ctx, err ? err : "Could not create new index");
+
+    RedisModule_ReplyWithError(ctx, err ? err : "Could not create new index");
+    ERR_FREE(err);
+    return REDISMODULE_OK;
   }
 
   return RedisModule_ReplyWithSimpleString(ctx, "OK");
@@ -1265,6 +1259,170 @@ int SuggestGetCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
   return REDISMODULE_OK;
 }
 
+/**
+ * FT.SYNADD <index> <term1> <term2> ...
+ *
+ * Add a synonym group to the given index. The synonym data structure is compose of synonyms groups.
+ * Each Synonym group has a unique id. The SYNADD command creates a new synonym group with the given
+ * terms and return its id.
+ */
+int SynAddCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
+  if (argc < 3) return RedisModule_WrongArity(ctx);
+
+  IndexSpec *sp = IndexSpec_Load(ctx, RedisModule_StringPtrLen(argv[1], NULL), 0);
+  if (!sp) {
+    RedisModule_ReplyWithError(ctx, "Unknown index name");
+    return REDISMODULE_OK;
+  }
+
+  IndexSpec_InitializeSynonym(sp);
+
+  uint32_t id = SynonymMap_AddRedisStr(sp->smap, argv + 2, argc - 2);
+
+  RedisModule_ReplyWithLongLong(ctx, id);
+
+  return REDISMODULE_OK;
+}
+
+int SynUpdateCommandInternal(RedisModuleCtx *ctx, RedisModuleString *indexName, long long id,
+                             RedisModuleString **synonyms, size_t size, bool checkIdSanity) {
+  IndexSpec *sp = IndexSpec_Load(ctx, RedisModule_StringPtrLen(indexName, NULL), 0);
+  if (!sp) {
+    RedisModule_ReplyWithError(ctx, "Unknown index name");
+    return REDISMODULE_OK;
+  }
+
+  if (checkIdSanity && (!sp->smap || id >= SynonymMap_GetMaxId(sp->smap))) {
+    RedisModule_ReplyWithError(ctx, "given id does not exists");
+    return REDISMODULE_OK;
+  }
+
+  IndexSpec_InitializeSynonym(sp);
+
+  SynonymMap_UpdateRedisStr(sp->smap, synonyms, size, id);
+
+  RedisModule_ReplyWithSimpleString(ctx, "OK");
+
+  return REDISMODULE_OK;
+}
+
+/**
+ * FT.SYNUPDATE <index> <id> <term1> <term2> ...
+ *
+ * Update an already existing synonym group with the given terms.
+ * Its only to add new terms to a synonym group.
+ * return true on success.
+ */
+int SynUpdateCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
+  if (argc < 4) return RedisModule_WrongArity(ctx);
+
+  long long id;
+  if (RedisModule_StringToLongLong(argv[2], &id) != REDISMODULE_OK) {
+    RedisModule_ReplyWithError(ctx, "wrong parameters, id is not an integer");
+    return REDISMODULE_OK;
+  }
+
+  if (id < 0 || id > UINT32_MAX) {
+    RedisModule_ReplyWithError(ctx, "wrong parameters, id out of range");
+    return REDISMODULE_OK;
+  }
+
+  return SynUpdateCommandInternal(ctx, argv[1], id, argv + 3, argc - 3, true);
+}
+
+int SynForceUpdateCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
+  if (argc < 4) return RedisModule_WrongArity(ctx);
+
+  long long id;
+  if (RedisModule_StringToLongLong(argv[2], &id) != REDISMODULE_OK) {
+    RedisModule_ReplyWithError(ctx, "wrong parameters, id is not an integer");
+    return REDISMODULE_OK;
+  }
+
+  if (id < 0 || id > UINT32_MAX) {
+    RedisModule_ReplyWithError(ctx, "wrong parameters, id out of range");
+    return REDISMODULE_OK;
+  }
+
+  return SynUpdateCommandInternal(ctx, argv[1], id, argv + 3, argc - 3, false);
+}
+
+/**
+ * FT.SYNDUMP <index>
+ *
+ * Dump the synonym data structure in the following format:
+ *    - term1
+ *        - id1
+ *        - id2
+ *    - term2
+ *        - id3
+ *    - term3
+ *        - id4
+ */
+int SynDumpCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
+  if (argc < 2) return RedisModule_WrongArity(ctx);
+
+  IndexSpec *sp = IndexSpec_Load(ctx, RedisModule_StringPtrLen(argv[1], NULL), 0);
+  if (!sp) {
+    RedisModule_ReplyWithError(ctx, "Unknown index name");
+    return REDISMODULE_OK;
+  }
+
+  if (!sp->smap) {
+    RedisModule_ReplyWithArray(ctx, 0);
+    return REDISMODULE_OK;
+  }
+
+  size_t size;
+  TermData **terms_data = SynonymMap_DumpAllTerms(sp->smap, &size);
+
+  RedisModule_ReplyWithArray(ctx, size * 2);
+
+  for (int i = 0; i < size; ++i) {
+    TermData *t_data = terms_data[i];
+    RedisModule_ReplyWithStringBuffer(ctx, t_data->term, strlen(t_data->term));
+    RedisModule_ReplyWithArray(ctx, array_len(t_data->ids));
+    for (size_t j = 0; j < array_len(t_data->ids); ++j) {
+      RedisModule_ReplyWithLongLong(ctx, t_data->ids[j]);
+    }
+  }
+
+  rm_free(terms_data);
+
+  return REDISMODULE_OK;
+}
+
+int AlterIndexCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
+  // Need at least <cmd> <index> <subcommand> <args...>
+  if (argc < 5) {
+    return RedisModule_WrongArity(ctx);
+  }
+  // I'd like to use CmdSchema, but want to avoid the ugly <N> <list of N> stuff..
+  if (!RMUtil_StringEqualsCaseC(argv[2], "SCHEMA") || !RMUtil_StringEqualsCaseC(argv[3], "ADD")) {
+    return RedisModule_ReplyWithError(ctx, "Unknown command");
+  }
+  IndexSpec *sp = IndexSpec_Load(ctx, RedisModule_StringPtrLen(argv[1], NULL), 1);
+  if (!sp) {
+    return RedisModule_ReplyWithError(ctx, "Unknown index name");
+  }
+
+  const int schemaOffset = 4;
+  char *err = NULL;
+
+  if (argc - schemaOffset == 0) {
+    return RedisModule_ReplyWithError(ctx, "No fields provided");
+  }
+
+  int rc = IndexSpec_AddFieldsRedisArgs(sp, argv + schemaOffset, argc - schemaOffset, &err);
+  if (!rc) {
+    RedisModule_ReplyWithError(ctx, err);
+    ERR_FREE(err);
+  } else {
+    RedisModule_ReplyWithSimpleString(ctx, "OK");
+  }
+  return REDISMODULE_OK;
+}
+
 #define RM_TRY(f, ...)                                                         \
   if (f(__VA_ARGS__) == REDISMODULE_ERR) {                                     \
     RedisModule_Log(ctx, "warning", "Could not run " #f "(" #__VA_ARGS__ ")"); \
@@ -1273,6 +1431,45 @@ int SuggestGetCommand(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
     RedisModule_Log(ctx, "verbose", "Successfully executed " #f);              \
   }
 
+/**
+ * Check if we can run under the current AOF configuration. Returns true
+ * or false
+ */
+static int validateAofSettings(RedisModuleCtx *ctx) {
+  int rc = 1;
+
+  if (RedisModule_GetContextFlags == NULL) {
+    RedisModule_Log(ctx, "warning",
+                    "Could not determine if AOF is in use. AOF Rewrite will crash!");
+    return 1;
+  }
+
+  if ((RedisModule_GetContextFlags(ctx) & REDISMODULE_CTX_FLAGS_AOF) == 0) {
+    // AOF disabled. All is OK, and no further checks needed
+    return rc;
+  }
+
+  // Can't exexcute commands on the loading context, so make a new one
+  RedisModuleCtx *confCtx = RedisModule_GetThreadSafeContext(NULL);
+  RedisModuleCallReply *reply =
+      RedisModule_Call(confCtx, "CONFIG", "cc", "GET", "aof-use-rdb-preamble");
+  assert(reply);
+  assert(RedisModule_CallReplyType(reply) == REDISMODULE_REPLY_ARRAY);
+  assert(RedisModule_CallReplyLength(reply) == 2);
+  const char *value =
+      RedisModule_CallReplyStringPtr(RedisModule_CallReplyArrayElement(reply, 1), NULL);
+
+  // I tried using strcasecmp, but it seems that the yes/no replies have a trailing
+  // embedded newline in them
+  if (tolower(*value) == 'n') {
+    RedisModule_Log(ctx, "warning", "FATAL: aof-use-rdb-preamble required if AOF is used!");
+    rc = 0;
+  }
+  RedisModule_FreeCallReply(reply);
+  RedisModule_FreeThreadSafeContext(confCtx);
+  return rc;
+}
+
 int RediSearch_InitModuleInternal(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
 
   // Check that redis supports thread safe context. RC3 or below doesn't
@@ -1283,19 +1480,19 @@ int RediSearch_InitModuleInternal(RedisModuleCtx *ctx, RedisModuleString **argv,
                     "\t\t\t\tRedis will exit now!");
     return REDISMODULE_ERR;
   }
-  RedisModule_Log(ctx, "verbose", "Loading RediSearch module!");
+
+  // Print version string!
+  RedisModule_Log(ctx, "notice", "RediSearch version %d.%d.%d (Git=%s)", REDISEARCH_VERSION_MAJOR,
+                  REDISEARCH_VERSION_MINOR, REDISEARCH_VERSION_PATCH, RS_GetExtraVersion());
 
   const char *err;
   if (ReadConfig(argv, argc, &err) == REDISMODULE_ERR) {
     RedisModule_Log(ctx, "warning", "Invalid Configurations: %s", err);
     return REDISMODULE_ERR;
   }
-  RedisModule_Log(ctx, "notice",
-                  "Configuration: concurrent mode: %d, ext load: %s, min prefix: %d, max "
-                  "expansions: %d, query timeout: %dms",
-                  RSGlobalConfig.concurrentMode, RSGlobalConfig.extLoad,
-                  RSGlobalConfig.minTermPrefix, RSGlobalConfig.maxPrefixExpansions,
-                  RSGlobalConfig.queryTimeoutMS);
+  sds confstr = RSConfig_GetInfoString(&RSGlobalConfig);
+  RedisModule_Log(ctx, "notice", confstr);
+  sdsfree(confstr);
 
   if (RedisModule_GetContextFlags == NULL && RSGlobalConfig.concurrentMode) {
     RedisModule_Log(ctx, "warning",
@@ -1304,10 +1501,21 @@ int RediSearch_InitModuleInternal(RedisModuleCtx *ctx, RedisModuleString **argv,
                     "malfunction unless 'safe' functions are used or SAFEMODE is enabled.");
   }
 
+  if (!validateAofSettings(ctx)) {
+    return REDISMODULE_ERR;
+  }
+
   // Init extension mechanism
   Extensions_Init();
 
   ConcurrentSearch_ThreadPoolStart();
+
+  // Init Schemata
+  Aggregate_BuildSchema();
+
+  // Init cursors mechanism
+  CursorList_Init(&RSCursors);
+
   RedisModule_Log(ctx, "notice", "Initialized thread pool!");
 
   /* Load extensions if needed */
@@ -1318,7 +1526,7 @@ int RediSearch_InitModuleInternal(RedisModuleCtx *ctx, RedisModuleString **argv,
     if (Extension_LoadDynamic(RSGlobalConfig.extLoad, &errMsg) == REDISMODULE_ERR) {
       RedisModule_Log(ctx, "warning", "Could not load extension %s: %s", RSGlobalConfig.extLoad,
                       errMsg);
-      free(errMsg);
+      ERR_FREE(errMsg);
       return REDISMODULE_ERR;
     }
     RedisModule_Log(ctx, "notice", "Loaded RediSearch extension '%s'", RSGlobalConfig.extLoad);
@@ -1354,13 +1562,10 @@ int RediSearch_InitModuleInternal(RedisModuleCtx *ctx, RedisModuleString **argv,
   RM_TRY(RedisModule_CreateCommand, ctx, RS_SAFEADDHASH_CMD, SafeAddHashCommand, "write deny-oom",
          1, 1, 1);
 
-  RM_TRY(RedisModule_CreateCommand, ctx, RS_DTADD_CMD, DTAddCommand, "write deny-oom", 1, 1, 1);
-
-  RM_TRY(RedisModule_CreateCommand, ctx, RS_ADDTERM_CMD, TermAddCommand, "write deny-oom", 1, 1, 1);
-
   RM_TRY(RedisModule_CreateCommand, ctx, RS_DEL_CMD, DeleteCommand, "write", 1, 1, 1);
 
   RM_TRY(RedisModule_CreateCommand, ctx, RS_SEARCH_CMD, SearchCommand, "readonly", 1, 1, 1);
+  RM_TRY(RedisModule_CreateCommand, ctx, RS_AGGREGATE_CMD, AggregateCommand, "readonly", 1, 1, 1);
 
   RM_TRY(RedisModule_CreateCommand, ctx, RS_GET_CMD, GetSingleDocumentCommand, "readonly", 1, 1, 1);
 
@@ -1387,18 +1592,20 @@ int RediSearch_InitModuleInternal(RedisModuleCtx *ctx, RedisModuleString **argv,
 
   RM_TRY(RedisModule_CreateCommand, ctx, RS_SUGGET_CMD, SuggestGetCommand, "readonly", 1, 1, 1);
 
+  RM_TRY(RedisModule_CreateCommand, ctx, RS_CURSOR_CMD, CursorCommand, "readonly", 2, 2, 1);
+
+  RM_TRY(RedisModule_CreateCommand, ctx, RS_SYNADD_CMD, SynAddCommand, "write", 1, 1, 1);
+
+  RM_TRY(RedisModule_CreateCommand, ctx, RS_SYNUPDATE_CMD, SynUpdateCommand, "write", 1, 1, 1);
+
+  RM_TRY(RedisModule_CreateCommand, ctx, RS_SYNFORCEUPDATE_CMD, SynForceUpdateCommand, "write", 1,
+         1, 1);
+
+  RM_TRY(RedisModule_CreateCommand, ctx, RS_SYNDUMP_CMD, SynDumpCommand, "readonly", 1, 1, 1);
+
+  RM_TRY(RedisModule_CreateCommand, ctx, RS_ALTER_CMD, AlterIndexCommand, "write", 1, 1, 1);
+
+  RM_TRY(RedisModule_CreateCommand, ctx, RS_DEBUG, DebugCommand, "readonly", 1, 1, 1);
+
   return REDISMODULE_OK;
 }
-
-/* Only compile the entry point if compiling the module as a standalone module */
-#ifndef RS_STATIC
-
-int RedisModule_OnLoad(RedisModuleCtx *ctx, RedisModuleString **argv, int argc) {
-
-  if (RedisModule_Init(ctx, "ft", REDISEARCH_MODULE_VERSION, REDISMODULE_APIVER_1) ==
-      REDISMODULE_ERR)
-    return REDISMODULE_ERR;
-  return RediSearch_InitModuleInternal(ctx, argv, argc);
-}
-
-#endif
