@@ -4,8 +4,10 @@
 #include "rmutil/vector.h"
 #include "rmutil/util.h"
 #include "index.h"
+#include "util/arr.h"
 #include <math.h>
 #include "redismodule.h"
+#include "util/misc.h"
 //#include "tests/time_sample.h"
 #define NR_EXPONENT 4
 #define NR_MAXRANGE_CARD 2500
@@ -68,7 +70,7 @@ int NumericRange_Add(NumericRange *n, t_docId docId, double value, int checkCard
   if (checkCard) {
     add = 1;
     size_t card = n->card;
-    for (int i = 0; i < card; i++) {
+    for (int i = 0; i < MIN(card, n->splitCard); i++) {
 
       if (n->values[i] == value) {
         add = 0;
@@ -81,6 +83,7 @@ int NumericRange_Add(NumericRange *n, t_docId docId, double value, int checkCard
   if (add) {
     if (n->card < n->splitCard) {
       n->values[n->card] = value;
+      n->unique_sum += value;
     }
     ++n->card;
   }
@@ -92,7 +95,7 @@ int NumericRange_Add(NumericRange *n, t_docId docId, double value, int checkCard
 
 double NumericRange_Split(NumericRange *n, NumericRangeNode **lp, NumericRangeNode **rp) {
 
-  double split = (n->minVal + n->maxVal) / (double)2;
+  double split = (n->unique_sum) / (double)n->card;
 
   // printf("split point :%f\n", split);
   *lp = NewLeafNode(n->entries->numDocs / 2 + 1, n->minVal, split,
@@ -127,6 +130,7 @@ NumericRangeNode *NewLeafNode(size_t cap, double min, double max, size_t splitCa
 
   *n->range = (NumericRange){.minVal = min,
                              .maxVal = max,
+                             .unique_sum = 0,
                              .card = 0,
                              .splitCard = splitCard,
                              .values = RedisModule_Calloc(splitCard, sizeof(double)),
@@ -134,22 +138,23 @@ NumericRangeNode *NewLeafNode(size_t cap, double min, double max, size_t splitCa
   return n;
 }
 
-#define __isLeaf(n) (n->left == NULL && n->right == NULL)
-
 int NumericRangeNode_Add(NumericRangeNode *n, t_docId docId, double value) {
 
-  if (!__isLeaf(n)) {
+  if (!NumericRangeNode_IsLeaf(n)) {
     // if this node has already split but retains a range, just add to the range without checking
     // anything
     if (n->range) {
       NumericRange_Add(n->range, docId, value, 0);
     }
 
-    // recursively add to its left or right child. if the child has split we get 1 in return
-    int rc = NumericRangeNode_Add((value < n->value ? n->left : n->right), docId, value);
+    // recursively add to its left or right child.
+    NumericRangeNode **childP = value < n->value ? &n->left : &n->right;
+    NumericRangeNode *child = *childP;
+    // if the child has split we get 1 in return
+    int rc = NumericRangeNode_Add(child, docId, value);
     if (rc) {
       // if there was a split it means our max depth has increased.
-      // we we are too deep - we don't retain this node's range anymore.
+      // we are too deep - we don't retain this node's range anymore.
       // this keeps memory footprint in check
       if (++n->maxDepth > NR_MAX_DEPTH && n->range) {
         InvertedIndex_Free(n->range->entries);
@@ -157,12 +162,30 @@ int NumericRangeNode_Add(NumericRangeNode *n, t_docId docId, double value) {
         RedisModule_Free(n->range);
         n->range = NULL;
       }
+
+      // check if we need to rebalance the child.
+      // To ease the rebalance we don't rebalance the root
+      // nor do we rebalance nodes that are with ranges (n->maxDepth > NR_MAX_DEPTH)
+      if ((child->right->maxDepth - child->left->maxDepth) > NR_MAX_DEPTH) {  // role to the left
+        NumericRangeNode *right = child->right;
+        child->right = right->left;
+        right->left = child;
+        --child->maxDepth;
+        *childP = right;  // replace the child with the new child
+      } else if ((child->left->maxDepth - child->right->maxDepth) >
+                 NR_MAX_DEPTH) {  // role to the right
+        NumericRangeNode *left = child->left;
+        child->left = left->right;
+        left->right = child;
+        --child->maxDepth;
+        *childP = left;  // replace the child with the new child
+      }
     }
     // return 1 or 0 to our called, so this is done recursively
     return rc;
   }
 
-  // if this node is a leaf - we add AND check the cardinlity. We only split leaf nodes
+  // if this node is a leaf - we add AND check the cardinality. We only split leaf nodes
   int card = NumericRange_Add(n->range, docId, value, 1);
 
   // printf("Added %d %f to node %f..%f, card now %zd, size now %zd\n", docId, value,
@@ -183,7 +206,7 @@ int NumericRangeNode_Add(NumericRangeNode *n, t_docId docId, double value) {
   return 0;
 }
 
-/* recrusively add a node's children to the range. */
+/* Recursively add a node's children to the range. */
 void __recursiveAddRange(Vector *v, NumericRangeNode *n, double min, double max) {
   if (!n) return;
 
@@ -204,7 +227,7 @@ void __recursiveAddRange(Vector *v, NumericRangeNode *n, double min, double max)
   }
 
   // for non leaf nodes - we try to descend into their children
-  if (!__isLeaf(n)) {
+  if (!NumericRangeNode_IsLeaf(n)) {
     __recursiveAddRange(v, n->left, min, max);
     __recursiveAddRange(v, n->right, min, max);
   } else if (NumericRange_Overlaps(n->range, min, max)) {
@@ -247,6 +270,7 @@ void NumericRangeNode_Free(NumericRangeNode *n) {
 
 /* Create a new numeric range tree */
 NumericRangeTree *NewNumericRangeTree() {
+#define GC_NODES_INITIAL_SIZE 10
   NumericRangeTree *ret = RedisModule_Alloc(sizeof(NumericRangeTree));
 
   ret->root = NewLeafNode(2, NF_NEGATIVE_INFINITY, NF_INFINITY, 2);
@@ -348,7 +372,7 @@ IndexIterator *createNumericIterator(NumericRangeTree *t, NumericFilter *f) {
   }
   Vector_Free(v);
 
-  IndexIterator *it = NewUnionIterator(its, n, NULL, 1);
+  IndexIterator *it = NewUnionIterator(its, n, NULL, 1, 1);
 
   return it;
 }
@@ -385,18 +409,16 @@ struct indexIterator *NewNumericFilterIterator(RedisSearchCtx *ctx, NumericFilte
   return it;
 }
 
-NumericRangeTree *OpenNumericIndex(RedisSearchCtx *ctx, const char *fname,
+NumericRangeTree *OpenNumericIndex(RedisSearchCtx *ctx, RedisModuleString *keyName,
                                    RedisModuleKey **idxKey) {
 
-  RedisModuleString *s = fmtRedisNumericIndexKey(ctx, fname);
   RedisModuleKey *key_s = NULL;
 
   if (!idxKey) {
     idxKey = &key_s;
   }
 
-  *idxKey = RedisModule_OpenKey(ctx->redisCtx, s, REDISMODULE_READ | REDISMODULE_WRITE);
-  RedisModule_FreeString(ctx->redisCtx, s);
+  *idxKey = RedisModule_OpenKey(ctx->redisCtx, keyName, REDISMODULE_READ | REDISMODULE_WRITE);
 
   int type = RedisModule_KeyType(*idxKey);
   if (type != REDISMODULE_KEYTYPE_EMPTY &&
@@ -435,16 +457,18 @@ unsigned long NumericIndexType_MemUsage(const void *value) {
   return ret;
 }
 
+#define NUMERIC_INDEX_ENCVER 1
+
 int NumericIndexType_Register(RedisModuleCtx *ctx) {
 
   RedisModuleTypeMethods tm = {.version = REDISMODULE_TYPE_METHOD_VERSION,
                                .rdb_load = NumericIndexType_RdbLoad,
                                .rdb_save = NumericIndexType_RdbSave,
-                               .aof_rewrite = NumericIndexType_AofRewrite,
+                               .aof_rewrite = GenericAofRewrite_DisabledHandler,
                                .free = NumericIndexType_Free,
                                .mem_usage = NumericIndexType_MemUsage};
 
-  NumericIndexType = RedisModule_CreateDataType(ctx, "numericdx", 0, &tm);
+  NumericIndexType = RedisModule_CreateDataType(ctx, "numericdx", NUMERIC_INDEX_ENCVER, &tm);
   if (NumericIndexType == NULL) {
     return REDISMODULE_ERR;
   }
@@ -459,50 +483,81 @@ typedef struct {
   double value;
 } NumericRangeEntry;
 
-static int __cmd_docId(const void *p1, const void *p2) {
+static int cmpdocId(const void *p1, const void *p2) {
   NumericRangeEntry *e1 = (NumericRangeEntry *)p1;
   NumericRangeEntry *e2 = (NumericRangeEntry *)p2;
 
   return (int)e1->docId - (int)e2->docId;
 }
-void *NumericIndexType_RdbLoad(RedisModuleIO *rdb, int encver) {
-  if (encver != 0) {
+
+/** Version 0 stores the number of entries beforehand, and then loads them */
+static size_t loadV0(RedisModuleIO *rdb, NumericRangeEntry **entriespp) {
+  uint64_t num = RedisModule_LoadUnsigned(rdb);
+  if (!num) {
     return 0;
   }
 
-  NumericRangeTree *t = NewNumericRangeTree();
-  uint64_t num = RedisModule_LoadUnsigned(rdb);
+  *entriespp = array_newlen(NumericRangeEntry, num);
+  NumericRangeEntry *entries = *entriespp;
+  for (size_t ii = 0; ii < num; ++ii) {
+    entries[ii].docId = RedisModule_LoadUnsigned(rdb);
+    entries[ii].value = RedisModule_LoadDouble(rdb);
+  }
+  return num;
+}
 
-  // we create an array of all the entries so that we can sort them by docId
-  NumericRangeEntry *entries = calloc(num, sizeof(NumericRangeEntry));
-  size_t n = 0;
-  for (size_t i = 0; i < num; i++) {
-    entries[n].docId = RedisModule_LoadUnsigned(rdb);
-    entries[n].value = RedisModule_LoadDouble(rdb);
-    n++;
+#define NUMERIC_IDX_INITIAL_LOAD_SIZE 1 << 16
+/** Version 0 stores (id,value) pairs, with a final 0 as a terminator */
+static size_t loadV1(RedisModuleIO *rdb, NumericRangeEntry **entriespp) {
+  NumericRangeEntry *entries = array_new(NumericRangeEntry, NUMERIC_IDX_INITIAL_LOAD_SIZE);
+  while (1) {
+    NumericRangeEntry cur;
+    cur.docId = RedisModule_LoadUnsigned(rdb);
+    if (!cur.docId) {
+      break;
+    }
+    cur.value = RedisModule_LoadDouble(rdb);
+    entries = array_append(entries, cur);
+  }
+  *entriespp = entries;
+  return array_len(entries);
+}
+
+void *NumericIndexType_RdbLoad(RedisModuleIO *rdb, int encver) {
+  if (encver > NUMERIC_INDEX_ENCVER) {
+    return NULL;
+  }
+
+  NumericRangeEntry *entries = NULL;
+  size_t numEntries = 0;
+  if (encver == 0) {
+    numEntries = loadV0(rdb, &entries);
+  } else if (encver == 1) {
+    numEntries = loadV1(rdb, &entries);
+  } else {
+    return NULL;  // Unknown version
   }
 
   // sort the entries by doc id, as they were not saved in this order
-  qsort(entries, num, sizeof(NumericRangeEntry), __cmd_docId);
+  qsort(entries, numEntries, sizeof(NumericRangeEntry), cmpdocId);
+  NumericRangeTree *t = NewNumericRangeTree();
 
   // now push them in order into the tree
-  for (size_t i = 0; i < num; i++) {
+  for (size_t i = 0; i < numEntries; i++) {
     NumericRangeTree_Add(t, entries[i].docId, entries[i].value);
   }
-  free(entries);
-
+  array_free(entries);
   return t;
 }
 
-struct __niRdbSaveCtx {
+struct niRdbSaveCtx {
   RedisModuleIO *rdb;
-  size_t num;
 };
 
-void __numericIndex_rdbSaveCallback(NumericRangeNode *n, void *ctx) {
-  struct __niRdbSaveCtx *rctx = ctx;
+static void numericIndex_rdbSaveCallback(NumericRangeNode *n, void *ctx) {
+  struct niRdbSaveCtx *rctx = ctx;
 
-  if (__isLeaf(n) && n->range) {
+  if (NumericRangeNode_IsLeaf(n) && n->range) {
     NumericRange *rng = n->range;
     RSIndexResult *res = NULL;
     IndexReader *ir = NewNumericReader(rng->entries, NULL);
@@ -517,16 +572,11 @@ void __numericIndex_rdbSaveCallback(NumericRangeNode *n, void *ctx) {
 void NumericIndexType_RdbSave(RedisModuleIO *rdb, void *value) {
 
   NumericRangeTree *t = value;
+  struct niRdbSaveCtx ctx = {rdb};
 
-  RedisModule_SaveUnsigned(rdb, t->numEntries);
-
-  struct __niRdbSaveCtx ctx = {rdb, 0};
-
-  NumericRangeNode_Traverse(t->root, __numericIndex_rdbSaveCallback, &ctx);
-}
-
-void NumericIndexType_AofRewrite(RedisModuleIO *aof, RedisModuleString *key, void *value) {
-  RMUtil_DefaultAofRewrite(aof, key, value);
+  NumericRangeNode_Traverse(t->root, numericIndex_rdbSaveCallback, &ctx);
+  // Save the final record
+  RedisModule_SaveUnsigned(rdb, 0);
 }
 
 void NumericIndexType_Digest(RedisModuleDigest *digest, void *value) {
@@ -536,3 +586,29 @@ void NumericIndexType_Free(void *value) {
   NumericRangeTree *t = value;
   NumericRangeTree_Free(t);
 }
+
+NumericRangeTreeIterator *NumericRangeTreeIterator_New(NumericRangeTree *t) {
+#define NODE_STACK_INITIAL_SIZE 4
+  NumericRangeTreeIterator *iter = rm_malloc(sizeof(NumericRangeTreeIterator));
+  iter->nodesStack = array_new(NumericRangeNode *, NODE_STACK_INITIAL_SIZE);
+  array_append(iter->nodesStack, t->root);
+  return iter;
+}
+
+NumericRangeNode *NumericRangeTreeIterator_Next(NumericRangeTreeIterator *iter) {
+  if (array_len(iter->nodesStack) == 0) {
+    return NULL;
+  }
+  NumericRangeNode *ret = array_pop(iter->nodesStack);
+  if (!NumericRangeNode_IsLeaf(ret)) {
+    iter->nodesStack = array_append(iter->nodesStack, ret->left);
+    iter->nodesStack = array_append(iter->nodesStack, ret->right);
+  }
+
+  return ret;
+}
+
+void NumericRangeTreeIterator_Free(NumericRangeTreeIterator *iter) {
+  array_free(iter->nodesStack);
+  rm_free(iter);
+}
