@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-from rmtest import ModuleTestCase
+from rmtest import BaseModuleTestCase
 import redis
 import unittest
 from hotels import hotels
@@ -8,7 +8,10 @@ import random
 import time
 
 
-class SearchTestCase(ModuleTestCase('../redisearch.so')):
+class SearchTestCase(BaseModuleTestCase):
+    def setUp(self):
+        super(SearchTestCase, self).setUp()
+        self.client.flushdb()
 
     def testAdd(self):
         with self.redis() as r:
@@ -26,6 +29,72 @@ class SearchTestCase(ModuleTestCase('../redisearch.so')):
                 self.assertExists(r, prefix + ':idx/world')
                 self.assertExists(r, prefix + ':idx/lorem')
 
+    def testConditionalUpdate(self):
+        self.assertOk(self.cmd(
+            'ft.create', 'idx', 'schema', 'foo', 'text', 'bar', 'numeric', 'sortable'))
+        self.assertOk(self.cmd('ft.add', 'idx', '1', '1',
+                               'fields', 'foo', 'hello', 'bar', '123'))
+        self.assertOk(self.cmd('ft.add', 'idx', '1', '1', 'replace', 'if',
+                               '@foo == "hello"', 'fields', 'foo', 'world', 'bar', '123'))
+        self.assertEqual('NOADD', self.cmd('ft.add', 'idx', '1', '1', 'replace',
+                                           'if', '@foo == "hello"', 'fields', 'foo', 'world', 'bar', '123'))
+        self.assertEqual('NOADD', self.cmd('ft.add', 'idx', '1', '1', 'replace',
+                                           'if', '1 == 2', 'fields', 'foo', 'world', 'bar', '123'))
+        self.assertOk(self.cmd('ft.add', 'idx', '1', '1', 'replace', 'partial', 'if',
+                               '@foo == "world"', 'fields', 'bar', '234'))
+        self.assertOk(self.cmd('ft.add', 'idx', '1', '1', 'replace', 'if',
+                               '@bar == 234', 'fields', 'foo', 'hello', 'bar', '123'))
+        
+        # Ensure that conditionals are ignored if the document doesn't exist
+        self.assertOk(self.cmd('FT.ADD', 'idx', '666', '1', 'IF', '@bar > 42', 'FIELDS', 'bar', '15'))
+        # Ensure that it fails if we try again, because it already exists
+        self.assertEqual('NOADD', self.cmd('FT.ADD', 'idx', '666', '1', 'REPLACE', 'IF', '@bar > 42', 'FIELDS', 'bar', '15'))
+        # Ensure that it fails because we're not using 'REPLACE'
+        with self.assertResponseError():
+            self.assertOk(self.cmd('FT.ADD', 'idx', '666', '1', 'IF', '@bar > 42', 'FIELDS', 'bar', '15'))
+
+    def testUnionIdList(self):
+        """
+        Regression test for https://github.com/RedisLabsModules/RediSearch/issues/306
+        """
+        with self.redis() as r:
+            r.flushdb()
+            N = 100
+            self.assertOk(r.execute_command(
+                "ft.create", "test", "SCHEMA",  "tags", "TAG", "waypoint", "GEO"))
+            self.assertOk(r.execute_command(
+                "ft.add", "test", "1", "1", "FIELDS", "tags", "alberta", "waypoint", "-113.524,53.5244"))
+            self.assertOk(r.execute_command(
+                "ft.add", "test", "2", "1", "FIELDS", "tags", "ontario", "waypoint", "-79.395,43.661667"))
+            res = r.execute_command(
+                'ft.search', 'test', "@waypoint:[-113.52 53.52 20 mi]|@tags:{ontario}", 'nocontent')
+            self.assertEqual(res, [2, '2', '1'])
+
+    def testAttributes(self):
+        self.assertOk(self.cmd('ft.create', 'idx', 'schema',
+                               'title', 'text', 'body', 'text'))
+        self.assertOk(self.cmd('ft.add', 'idx', 'doc1', 1.0, 'fields',
+                                                'title', 't1 t2', 'body', 't3 t4 t5'))
+        self.assertOk(self.cmd('ft.add', 'idx', 'doc2', 1.0, 'fields',
+                               'body', 't1 t2', 'title', 't3 t5'))
+
+        res = self.cmd(
+            'ft.search', 'idx', '(@title:(t1 t2) => {$weight: 0.2}) |(@body:(t1 t2) => {$weight: 0.5})', 'nocontent')
+        self.assertListEqual([2L, 'doc2', 'doc1'], res)
+        res = self.cmd(
+            'ft.search', 'idx', '(@title:(t1 t2) => {$weight: 2.5}) |(@body:(t1 t2) => {$weight: 0.5})', 'nocontent')
+        self.assertListEqual([2L, 'doc1', 'doc2'], res)
+
+        res = self.cmd(
+            'ft.search', 'idx', '(t3 t5) => {$slop: 4}', 'nocontent')
+        self.assertListEqual([2L, 'doc2', 'doc1'], res)
+        res = self.cmd(
+            'ft.search', 'idx', '(t5 t3) => {$slop: 0}', 'nocontent')
+        self.assertListEqual([1L, 'doc2'], res)
+        res = self.cmd(
+            'ft.search', 'idx', '(t5 t3) => {$slop: 0; $inorder:true}', 'nocontent')
+        self.assertListEqual([0], res)
+
     def testUnion(self):
 
         with self.redis() as r:
@@ -1429,98 +1498,6 @@ class SearchTestCase(ModuleTestCase('../redisearch.so')):
                 for option in filter(None, combo):
                     self.assertTrue(option in opts)
 
-    def aofTestCommon(self, reloadfn):
-        self.spawn_server(use_aof=True)
-
-        # TODO: Change this attribute in rmtest
-        if self.server._is_external:
-            raise unittest.SkipTest('Cannot run AOF tests on external server')
-
-        self.cmd('ft.create', 'idx', 'schema',
-                 'field1', 'text', 'field2', 'numeric')
-        reloadfn()
-        for x in range(1, 10):
-            self.assertCmdOk('ft.add', 'idx', 'doc{}'.format(x), 1.0 / x, 'fields',
-                             'field1', 'myText{}'.format(x), 'field2', 20 * x)
-        exp = [9L, 'doc1', ['field1', 'myText1', 'field2', '20'], 'doc2', ['field1', 'myText2', 'field2', '40'], 'doc3', ['field1', 'myText3', 'field2', '60'], 'doc4', ['field1', 'myText4', 'field2', '80'], 'doc5', ['field1',
-                                                                                                                                                                                                                        'myText5', 'field2', '100'], 'doc6', ['field1', 'myText6', 'field2', '120'], 'doc7', ['field1', 'myText7', 'field2', '140'], 'doc8', ['field1', 'myText8', 'field2', '160'], 'doc9', ['field1', 'myText9', 'field2', '180']]
-        reloadfn()
-        ret = self.cmd('ft.search', 'idx', 'myt*')
-        self.assertEqual(exp, ret)
-
-    def testAof(self):
-        self.aofTestCommon(lambda: self.restart_and_reload())
-
-    def testRawAof(self):
-        self.aofTestCommon(lambda: self.cmd('debug loadaof'))
-
-    def testRewriteAofSortables(self):
-        self.spawn_server(use_aof=True)
-        self.cmd('FT.CREATE', 'idx', 'schema', 'field1', 'TEXT',
-                 'SORTABLE', 'num1', 'NUMERIC', 'SORTABLE')
-        self.cmd('FT.ADD', 'idx', 'doc', 1.0,
-                 'FIELDS', 'field1', 'Hello World')
-        self.restart_and_reload()
-        self.cmd('SAVE')
-
-        from random import randint
-
-        # Load some documents
-        for x in xrange(100):
-            self.cmd('FT.ADD', 'idx', 'doc{}'.format(x), 1.0, 'FIELDS',
-                     'field1', 'txt{}'.format(random.random()),
-                     'num1', random.random())
-
-        cmd = ['FT.SEARCH', 'idx', 'txt', 'SORTBY',
-               '2', 'field1', 'ASC', 'num1', 'DESC']
-        res = self.cmd(*cmd)
-        self.restart_and_reload()
-        res2 = self.cmd(*cmd)
-        self.assertEqual(res, res2)
-
-    def testAofRewriteSortkeys(self):
-        self.spawn_server(use_aof=True)
-        self.cmd('FT.CREATE', 'idx', 'SCHEMA', 'foo',
-                 'TEXT', 'SORTABLE', 'bar', 'TAG')
-        self.cmd('FT.ADD', 'idx', '1', '1', 'FIELDS', 'foo', 'A', 'bar', '1')
-        self.cmd('FT.ADD', 'idx', '2', '1', 'fields', 'foo', 'B', 'bar', '1')
-
-        res_exp = self.cmd('FT.SEARCH', 'idx', '@bar:{1}', 'SORTBY', 'foo', 'ASC',
-                           'RETURN', '1', 'foo', 'WITHSORTKEYS')
-
-        self.restart_and_reload()
-        res_got = self.cmd('FT.SEARCH', 'idx', '@bar:{1}', 'SORTBY', 'foo', 'ASC',
-                           'RETURN', '1', 'foo', 'WITHSORTKEYS')
-
-        self.assertEqual(res_exp, res_got)
-
-    def testAofRewriteTags(self):
-        self.spawn_server(use_aof=True)
-        self.cmd('FT.CREATE', 'idx', 'SCHEMA', 'foo',
-                 'TEXT', 'SORTABLE', 'bar', 'TAG')
-        self.cmd('FT.ADD', 'idx', '1', '1', 'FIELDS', 'foo', 'A', 'bar', '1')
-        self.cmd('FT.ADD', 'idx', '2', '1', 'fields', 'foo', 'B', 'bar', '1')
-
-        def to_dict(r):
-            return {r[i]: r[i + 1] for i in range(0, len(r), 2)}
-        info_a = to_dict(self.cmd('FT.INFO', 'idx'))
-        self.restart_and_reload()
-        info_b = to_dict(self.cmd('FT.INFO', 'idx'))
-        self.assertEqual(info_a['fields'], info_b['fields'])
-
-        # Try to drop the schema
-        self.cmd('FT.DROP', 'idx')
-
-        # Try to create it again - should work!
-        self.cmd('FT.CREATE', 'idx', 'SCHEMA', 'foo',
-                 'TEXT', 'SORTABLE', 'bar', 'TAG')
-        self.cmd('FT.ADD', 'idx', '1', '1', 'FIELDS', 'foo', 'A', 'bar', '1')
-        self.cmd('FT.ADD', 'idx', '2', '1', 'fields', 'foo', 'B', 'bar', '1')
-        res = self.cmd('FT.SEARCH', 'idx', '@bar:{1}', 'SORTBY', 'foo', 'ASC',
-                       'RETURN', '1', 'foo', 'WITHSORTKEYS')
-        self.assertEqual([2L, '1', '$a', ['foo', 'A'],
-                          '2', '$b', ['foo', 'B']], res)
-
     def testNoStem(self):
         self.cmd('ft.create', 'idx', 'schema', 'body',
                  'text', 'name', 'text', 'nostem')
@@ -1743,6 +1720,113 @@ class SearchTestCase(ModuleTestCase('../redisearch.so')):
             self.cmd('FT.ADD', 'idx', 'doc1', 1.0, 'LANGUAGE',
                      'blah', 'FIELDS', 'language', 'gibber')
 
+    def testUninitSortvector(self):
+        # This would previously crash
+        self.cmd('FT.CREATE', 'idx', 'SCHEMA', 'f1', 'TEXT')
+        for x in range(2000):
+            self.cmd('FT.ADD', 'idx', 'doc{}'.format(
+                x), 1.0, 'FIELDS', 'f1', 'HELLO')
+
+        self.cmd('SAVE')
+        for x in range(10):
+            self.cmd('DEBUG RELOAD')
+
+    def testAlterIndex(self):
+        self.cmd('FT.CREATE', 'idx', 'SCHEMA', 'f1', 'TEXT')
+        self.cmd('FT.ADD', 'idx', 'doc1', 1.0, 'FIELDS', 'f1', 'hello', 'f2', 'world')
+        self.cmd('FT.ALTER', 'idx', 'SCHEMA', 'ADD', 'f2', 'TEXT')
+        self.cmd('FT.ADD', 'idx', 'doc2', 1.0, 'FIELDS', 'f1', 'hello', 'f2', 'world')
+        for _ in self.retry_with_reload():
+            ret = self.cmd('FT.SEARCH', 'idx', 'world')
+            self.assertEqual([1, 'doc2', ['f1', 'hello', 'f2', 'world']], ret)
+
+        self.cmd('FT.ALTER', 'idx', 'SCHEMA', 'ADD', 'f3', 'TEXT', 'SORTABLE')
+        for x in range(10):
+            self.cmd('FT.ADD', 'idx', 'doc{}'.format(x + 3), 1.0, 'FIELDS', 'f1', 'hello', 'f3', 'val{}'.format(x))
+        
+        for _ in self.retry_with_reload():
+            # Test that sortable works
+            res = self.cmd('FT.SEARCH', 'idx', 'hello', 'SORTBY', 'f3', 'DESC')
+            self.assertEqual(
+                [12, 'doc12', ['f1', 'hello', 'f3', 'val9'], 'doc11', ['f1', 'hello', 'f3', 'val8'], 'doc10', ['f1', 'hello', 'f3', 'val7'], 'doc9', ['f1', 'hello', 'f3', 'val6'], 'doc8', ['f1', 'hello', 'f3', 'val5'], 'doc7', ['f1', 'hello', 'f3', 'val4'], 'doc6', ['f1', 'hello', 'f3', 'val3'], 'doc5', ['f1', 'hello', 'f3', 'val2'], 'doc4', ['f1', 'hello', 'f3', 'val1'], 'doc3', ['f1', 'hello', 'f3', 'val0']],
+                res)
+
+        # Test that we can add a numeric field
+        self.cmd('FT.ALTER', 'idx', 'SCHEMA', 'ADD', 'n1', 'NUMERIC')
+        self.cmd('FT.ADD', 'idx', 'docN1', 1.0, 'FIELDS', 'n1', 50)
+        self.cmd('FT.ADD', 'idx', 'docN2', 1.0, 'FIELDS', 'n1', 250)
+        for _ in self.retry_with_reload():
+            res = self.cmd('FT.SEARCH', 'idx', '@n1:[0 100]')
+            self.assertEqual([1, 'docN1', ['n1', '50']], res)
+
+    def testAlterValidation(self):
+        # Test that constraints for ALTER comand
+        self.cmd('FT.CREATE', 'idx1', 'SCHEMA', 'f0', 'TEXT')
+        for x in range(1, 32):
+            self.cmd('FT.ALTER', 'idx1', 'SCHEMA', 'ADD', 'f{}'.format(x), 'TEXT')
+        # OK for now.
+
+        # Should be too many indexes
+        self.assertRaises(redis.ResponseError, self.cmd, 'FT.ALTER', 'idx1', 'SCHEMA', 'ADD', 'tooBig', 'TEXT')
+
+        self.cmd('FT.CREATE', 'idx2', 'MAXTEXTFIELDS', 'SCHEMA', 'f0', 'TEXT')
+        # print self.cmd('FT.INFO', 'idx2')
+        for x in range(1, 50):
+            self.cmd('FT.ALTER', 'idx2', 'SCHEMA', 'ADD', 'f{}'.format(x + 1), 'TEXT')
+        
+        self.cmd('FT.ADD', 'idx2', 'doc1', 1.0, 'FIELDS', 'f50', 'hello')
+        for _ in self.retry_with_reload():
+            ret = self.cmd('FT.SEARCH', 'idx2', '@f50:hello')
+            self.assertEqual([1, 'doc1', ['f50', 'hello']], ret)
+        
+
+        self.cmd('FT.CREATE', 'idx3', 'SCHEMA', 'f0', 'text')
+        # Try to alter the index with garbage
+        self.assertRaises(redis.ResponseError, self.cmd, 'FT.ALTER', 'idx3', 'SCHEMA', 'ADD', 'f1', 'TEXT', 'f2', 'garbage')
+        ret = to_dict(self.cmd('ft.info', 'idx3'))
+        self.assertEqual(1, len(ret['fields']))
+
+        self.assertRaises(redis.ResponseError, self.cmd, 'FT.ALTER', 'nonExist', 'SCHEMA', 'ADD', 'f1', 'TEXT')
+
+        # test with no fields!
+        self.assertRaises(redis.ResponseError, self.cmd, 'FT.ALTER', 'idx2', 'SCHEMA', 'ADD')
+
+
+    def testIssue366(self):
+        # Test random RDB regressions, see GH 366
+        self.cmd('FT.CREATE', 'idx1', 'SCHEMA', 'textfield', 'TEXT', 'numfield', 'NUMERIC')
+        self.cmd('HSET', 'foo', 'textfield', 'blah', 'numfield', 1)
+        self.cmd('FT.ADDHASH', 'idx1', 'foo', 1, 'replace')
+        self.cmd('FT.DEL', 'idx1', 'foo')
+        for _ in self.retry_with_reload():
+            pass  #  --just ensure it doesn't crash
+        
+# FT.CREATE atest SCHEMA textfield TEXT numfield NUMERIC
+# FT.ADD atest anId 1 PAYLOAD '{"hello":"world"}' FIELDS textfield sometext numfield 1234
+# FT.ADD atest anId 1 PAYLOAD '{"hello":"world2"}' REPLACE PARTIAL FIELDS numfield 1111
+# shutdown
+        self.cmd('FT.CREATE', 'idx2', 'SCHEMA', 'textfield', 'TEXT', 'numfield', 'NUMERIC')
+        self.cmd('FT.ADD', 'idx1', 'doc1', 1, 'PAYLOAD', '{"hello":"world"}',
+            'FIELDS', 'textfield', 'sometext', 'numfield', 1234)
+        self.cmd('ft.add', 'idx1', 'doc1', 1,
+            'PAYLOAD', '{"hello":"world2"}',
+            'REPLACE', 'PARTIAL',
+            'FIELDS', 'textfield', 'sometext', 'numfield', 1111)
+        for _ in self.retry_with_reload():
+            pass  #
+
+    def testReplaceReload(self):
+        self.cmd('FT.CREATE', 'idx2', 'SCHEMA', 'textfield', 'TEXT', 'numfield', 'NUMERIC')
+        # Create a document and then replace it.
+        self.cmd('FT.ADD', 'idx2', 'doc2', 1.0, 'FIELDS', 'textfield', 's1', 'numfield', 99)
+        self.cmd('FT.ADD', 'idx2', 'doc2', 1.0, 'REPLACE', 'PARTIAL', 'FIELDS', 'textfield', 's100', 'numfield', 990)
+        self.server.dump_and_reload()
+        # RDB Should still be fine
+
+        self.cmd('FT.ADD', 'idx2', 'doc2', 1.0, 'REPLACE', 'PARTIAL', 'FIELDS', 'textfield', 's200', 'numfield', 1090)
+        doc = to_dict(self.cmd('FT.GET', 'idx2', 'doc2'))
+        self.assertEqual('s200', doc['textfield'])
+        self.assertEqual('1090', doc['numfield'])
 
 def grouper(iterable, n, fillvalue=None):
     "Collect data into fixed-length chunks or blocks"
@@ -1751,6 +1835,8 @@ def grouper(iterable, n, fillvalue=None):
     args = [iter(iterable)] * n
     return izip_longest(fillvalue=fillvalue, *args)
 
+def to_dict(r):
+    return {r[i]: r[i + 1] for i in range(0, len(r), 2)}
 
 if __name__ == '__main__':
 
