@@ -1,9 +1,9 @@
 import math
 import unittest
-from rmtest import ModuleTestCase
+from rmtest import BaseModuleTestCase
 
 
-class ScorersTestCase(ModuleTestCase('../redisearch.so')):
+class ScorersTestCase(BaseModuleTestCase):
 
     def search(self, r, *args):
         return r.execute_command('ft.search', *args)
@@ -22,7 +22,6 @@ class ScorersTestCase(ModuleTestCase('../redisearch.so')):
             for i in range(16):
                 res = r.execute_command('ft.search', 'idx', '*', 'PAYLOAD', ('%x' % i) * 8,
                                         'SCORER', 'HAMMING', 'WITHSCORES', 'WITHPAYLOADS')
-
                 self.assertEqual(res[1], 'doc%d' % i)
                 self.assertEqual(res[2], '1')
                 # test with payload of different lenght
