@@ -20,6 +20,16 @@ typedef struct {
   (RedisSearchCtx) {               \
     .redisCtx = ctx, .spec = sp    \
   }
+
+#define SEARCH_CTX_SORTABLES(ctx) ((ctx && ctx->spec) ? ctx->spec->sortables : NULL)
+// Create a string context on the heap
 RedisSearchCtx *NewSearchCtx(RedisModuleCtx *ctx, RedisModuleString *indexName);
+RedisSearchCtx *NewSearchCtxDefault(RedisModuleCtx *ctx);
+
+RedisSearchCtx *SearchCtx_Refresh(RedisSearchCtx *sctx, RedisModuleString *keyName);
+
+// Same as above, only from c string (null terminated)
+RedisSearchCtx *NewSearchCtxC(RedisModuleCtx *ctx, const char *indexName);
+
 void SearchCtx_Free(RedisSearchCtx *sctx);
 #endif
