@@ -2,6 +2,7 @@
 #include "rmutil/util.h"
 #include "spec.h"
 #include "util/logging.h"
+#include "util/misc.h"
 #include "rmutil/vector.h"
 #include "trie/trie_type.h"
 #include <math.h>
@@ -9,8 +10,13 @@
 #include <assert.h>
 #include "rmalloc.h"
 #include "config.h"
+#include "cursor.h"
+#include "tag_index.h"
+
+void (*IndexSpec_OnCreate)(const IndexSpec *) = NULL;
 
 RedisModuleType *IndexSpecType;
+uint64_t spec_unique_ids = 0;
 
 /*
  * Get a field spec by field name. Case insensitive!
@@ -74,35 +80,41 @@ IndexSpec *IndexSpec_CreateNew(RedisModuleCtx *ctx, RedisModuleString **argv, in
                                char **err) {
   IndexSpec *sp = IndexSpec_ParseRedisArgs(ctx, argv[1], &argv[2], argc - 2, err);
   if (sp == NULL) {
-    if (!*err) *err = "Could not parse index spec";
+    SET_ERR(err, "Could not parse index spec");
     return NULL;
   }
 
-  RedisModuleKey *k =
-      RedisModule_OpenKey(ctx, RedisModule_CreateStringPrintf(ctx, INDEX_SPEC_KEY_FMT, sp->name),
-                          REDISMODULE_READ | REDISMODULE_WRITE);
+  RedisModuleString *keyString = RedisModule_CreateStringPrintf(ctx, INDEX_SPEC_KEY_FMT, sp->name);
+  RedisModuleKey *k = RedisModule_OpenKey(ctx, keyString, REDISMODULE_READ | REDISMODULE_WRITE);
 
   // check that the key is empty
   if (k == NULL || (RedisModule_KeyType(k) != REDISMODULE_KEYTYPE_EMPTY)) {
     if (RedisModule_ModuleTypeGetType(k) != IndexSpecType) {
-      *err = "Wrong type for index key";
+      SET_ERR(err, "Wrong type for index key");
     } else {
-      *err = "Index already exists. Drop it first!";
+      SET_ERR(err, "Index already exists. Drop it first!");
     }
     IndexSpec_Free(sp);
     return NULL;
   }
 
+  sp->unique_id = spec_unique_ids++;
   // Start the garbage collector
   IndexSpec_StartGC(ctx, sp, GC_DEFAULT_HZ);
 
+  CursorList_AddSpec(&RSCursors, sp->name, RSCURSORS_DEFAULT_CAPACITY);
+
   // set the value in redis
   RedisModule_ModuleTypeSetValue(k, IndexSpecType, sp);
 
+  if (IndexSpec_OnCreate) {
+    IndexSpec_OnCreate(sp);
+  }
+
   return sp;
 }
 
-int __findOffset(const char *arg, const char **argv, int argc) {
+static int findOffset(const char *arg, const char **argv, int argc) {
   for (int i = 0; i < argc; i++) {
     if (!strcasecmp(arg, argv[i])) {
       return i;
@@ -111,8 +123,8 @@ int __findOffset(const char *arg, const char **argv, int argc) {
   return -1;
 }
 
-int __argExists(const char *arg, const char **argv, int argc, int maxIdx) {
-  int idx = __findOffset(arg, argv, argc);
+static int argExists(const char *arg, const char **argv, int argc, int maxIdx) {
+  int idx = findOffset(arg, argv, argc);
   // printf("pos for %s: %d\n", arg, idx);
   return idx >= 0 && idx < maxIdx;
 }
@@ -127,7 +139,7 @@ char *strtolower(char *str) {
 }
 /* Parse a field definition from argv, at *offset. We advance offset as we progress.
  *  Returns 1 on successful parse, 0 otherwise */
-int __parseFieldSpec(const char **argv, int *offset, int argc, FieldSpec *sp, char **err) {
+static int parseFieldSpec(const char **argv, int *offset, int argc, FieldSpec *sp, char **err) {
 
   // if we're at the end - fail
   if (*offset >= argc) return 0;
@@ -189,7 +201,7 @@ int __parseFieldSpec(const char **argv, int *offset, int argc, FieldSpec *sp, ch
       if (strlen(argv[*offset]) == 1) {
         sp->tagOpts.separator = argv[*offset][0];
       } else {
-        *err = "Invalid separator, only 1 byte ascii characters allowed";
+        SET_ERR(err, "Invalid separator, only 1 byte ascii characters allowed");
         return 0;
       }
       ++*offset;
@@ -201,8 +213,8 @@ int __parseFieldSpec(const char **argv, int *offset, int argc, FieldSpec *sp, ch
   while (*offset < argc) {
     if (!strcasecmp(argv[*offset], SPEC_SORTABLE_STR)) {
       // cannot sort by geo fields
-      if (sp->type == FIELD_GEO || sp->type == FIELD_TAG) {
-        *err = "Tag and Geo fields cannot be sortable";
+      if (sp->type == FIELD_GEO) {
+        SET_ERR(err, "Geo fields cannot be sortable");
         return 0;
       }
       sp->options |= FieldSpec_Sortable;
@@ -218,14 +230,102 @@ int __parseFieldSpec(const char **argv, int *offset, int argc, FieldSpec *sp, ch
   return 1;
 }
 
-void _spec_buildSortingTable(IndexSpec *spec, int len) {
-  spec->sortables = NewSortingTable(len);
-  for (int i = 0; i < spec->numFields; i++) {
-    if (FieldSpec_IsSortable(&spec->fields[i])) {
-      // printf("Adding sortable field %s id %d\n", spec->fields[i].name, spec->fields[i].sortIdx);
-      SortingTable_SetFieldName(spec->sortables, spec->fields[i].sortIdx, spec->fields[i].name);
+/* Convert field type rsvalue type */
+RSValueType fieldTypeToValueType(FieldType ft) {
+  switch (ft) {
+    case FIELD_NUMERIC:
+      return RSValue_Number;
+    case FIELD_FULLTEXT:
+    case FIELD_TAG:
+      return RSValue_String;
+    case FIELD_GEO:
+      // geo is not sortable so we don't care as of now...
+      return RSValue_Null;
+  }
+}
+
+/**
+ * Add fields to an existing (or newly created) index. If the addition fails,
+ *
+ */
+static int IndexSpec_AddFieldsInternal(IndexSpec *sp, const char **argv, int argc, char **err,
+                                       int isNew) {
+
+  const size_t prevNumFields = sp->numFields;
+  const size_t prevSortLen = sp->sortables->len;
+
+  int textId = -1;
+
+  for (size_t ii = 0; ii < sp->numFields; ++ii) {
+    const FieldSpec *fs = sp->fields + ii;
+    if (fs->type == FIELD_FULLTEXT) {
+      textId = MAX(textId, fs->textOpts.id);
     }
   }
+
+  for (int offset = 0; offset < argc && sp->numFields < SPEC_MAX_FIELDS;) {
+    sp->fields = rm_realloc(sp->fields, sizeof(*sp->fields) * (sp->numFields + 1));
+    FieldSpec *fs = sp->fields + sp->numFields;
+    memset(fs, 0, sizeof(*fs));
+
+    fs->index = sp->numFields;
+
+    if (!parseFieldSpec(argv, &offset, argc, fs, err)) {
+      SET_ERR(err, "Could not parse field spec");
+      goto reset;
+    }
+
+    if (fs->type == FIELD_FULLTEXT && FieldSpec_IsIndexable(fs)) {
+      // make sure we don't have too many indexable fields
+      textId++;  // Explicit
+      if (textId == SPEC_MAX_FIELD_ID) {
+        SET_ERR(err, "Too many TEXT fields in schema");
+        goto reset;
+      }
+
+      // If we need to store field flags and we have over 32 fields, we need to switch to wide
+      // schema encoding
+      if (textId >= SPEC_WIDEFIELD_THRESHOLD && (sp->flags & Index_StoreFieldFlags)) {
+        if (isNew) {
+          sp->flags |= Index_WideSchema;
+        } else if ((sp->flags & Index_WideSchema) == 0) {
+          SET_ERR(err,
+                  "Cannot add more fields. Declare index with wide fields to allow adding "
+                  "unlimited fields");
+          goto reset;
+        }
+      }
+      fs->textOpts.id = textId;
+    }
+
+    if (IndexSpec_GetField(sp, fs->name, strlen(fs->name))) {
+      SET_ERR(err, "Duplicate field in schema");
+      goto reset;
+    }
+
+    if (FieldSpec_IsSortable(fs)) {
+      fs->sortIdx = RSSortingTable_Add(sp->sortables, fs->name, fieldTypeToValueType(fs->type));
+    }
+    sp->numFields++;
+  }
+  return 1;
+
+reset:
+  sp->numFields = prevNumFields;
+  sp->sortables->len = prevSortLen;
+  return 0;
+}
+
+int IndexSpec_AddFields(IndexSpec *sp, const char **argv, int argc, char **err) {
+  return IndexSpec_AddFieldsInternal(sp, argv, argc, err, 0);
+}
+
+int IndexSpec_AddFieldsRedisArgs(IndexSpec *sp, RedisModuleString **argv, int argc, char **err) {
+  const char *args[argc];
+  for (int i = 0; i < argc; i++) {
+    args[i] = RedisModule_StringPtrLen(argv[i], NULL);
+  }
+  return IndexSpec_AddFields(sp, args, argc, err);
 }
 
 /* The format currently is FT.CREATE {index} [NOOFFSETS] [NOFIELDS]
@@ -233,35 +333,39 @@ void _spec_buildSortingTable(IndexSpec *spec, int len) {
   */
 IndexSpec *IndexSpec_Parse(const char *name, const char **argv, int argc, char **err) {
   *err = NULL;
-  int schemaOffset = __findOffset(SPEC_SCHEMA_STR, argv, argc);
+  int schemaOffset = findOffset(SPEC_SCHEMA_STR, argv, argc);
   // no schema or schema towrards the end
   if (schemaOffset == -1) {
-    *err = "schema not found";
+    SET_ERR(err, "schema not found");
     return NULL;
   }
   IndexSpec *spec = NewIndexSpec(name, 0);
 
-  if (__argExists(SPEC_NOOFFSETS_STR, argv, argc, schemaOffset)) {
+  if (argExists(SPEC_NOOFFSETS_STR, argv, argc, schemaOffset)) {
     spec->flags &= ~(Index_StoreTermOffsets | Index_StoreByteOffsets);
   }
 
-  if (__argExists(SPEC_NOHL_STR, argv, argc, schemaOffset)) {
+  if (argExists(SPEC_NOHL_STR, argv, argc, schemaOffset)) {
     spec->flags &= ~Index_StoreByteOffsets;
   }
 
-  if (__argExists(SPEC_NOFIELDS_STR, argv, argc, schemaOffset)) {
+  if (argExists(SPEC_NOFIELDS_STR, argv, argc, schemaOffset)) {
     spec->flags &= ~Index_StoreFieldFlags;
   }
 
-  if (__argExists(SPEC_NOFREQS_STR, argv, argc, schemaOffset)) {
+  if (argExists(SPEC_NOFREQS_STR, argv, argc, schemaOffset)) {
     spec->flags &= ~Index_StoreFreqs;
   }
 
-  int swIndex = __findOffset(SPEC_STOPWORDS_STR, argv, argc);
+  if (argExists(SPEC_SCHEMA_EXPANDABLE_STR, argv, argc, schemaOffset)) {
+    spec->flags |= Index_WideSchema;
+  }
+
+  int swIndex = findOffset(SPEC_STOPWORDS_STR, argv, argc);
   if (swIndex >= 0 && swIndex + 1 < schemaOffset) {
     int listSize = atoi(argv[swIndex + 1]);
     if (listSize < 0 || (swIndex + 2 + listSize > schemaOffset)) {
-      *err = "Invalid stopword list size";
+      SET_ERR(err, "Invalid stopword list size");
       goto failure;
     }
     spec->stopwords = NewStopWordListCStr(&argv[swIndex + 2], listSize);
@@ -269,58 +373,10 @@ IndexSpec *IndexSpec_Parse(const char *name, const char **argv, int argc, char *
   } else {
     spec->stopwords = DefaultStopWordList();
   }
-
-  uint64_t id = 0;
-  int sortIdx = 0;
-
-  int i = schemaOffset + 1;
-  while (i < argc && spec->numFields < SPEC_MAX_FIELDS) {
-
-    FieldSpec *fs = &spec->fields[spec->numFields++];
-    fs->index = spec->numFields - 1;
-
-    if (!__parseFieldSpec(argv, &i, argc, fs, err)) {
-      if (!*err) {
-        *err = "Could not parse field spec";
-      }
-      goto failure;
-    }
-
-    if (fs->type == FIELD_FULLTEXT && FieldSpec_IsIndexable(fs)) {
-      // make sure we don't have too many indexable fields
-      if (id == SPEC_MAX_FIELD_ID) {
-        *err = "Too many TEXT fields in schema";
-        goto failure;
-      }
-
-      // If we need to store field flags and we have over 32 fields, we need to switch to wide
-      // schema encoding
-      if (id == SPEC_WIDEFIELD_THRESHOLD && spec->flags & Index_StoreFieldFlags) {
-        spec->flags |= Index_WideSchema;
-      }
-
-      fs->textOpts.id = id++;
-    }
-    if (FieldSpec_IsSortable(fs)) {
-      fs->sortIdx = sortIdx++;
-    }
-    if (sortIdx > RS_SORTABLES_MAX) {
-      *err = "Too many sortable fields";
-
-      goto failure;
-    }
-
-    if (IndexSpec_GetField(spec, fs->name, strlen(fs->name)) != fs) {
-      *err = "Duplicate field in schema";
-      goto failure;
-    }
+  schemaOffset++;
+  if (!IndexSpec_AddFieldsInternal(spec, argv + schemaOffset, argc - schemaOffset, err, 1)) {
+    goto failure;
   }
-
-  /* If we have sortable fields, create a sorting lookup table */
-  if (sortIdx > 0) {
-    _spec_buildSortingTable(spec, sortIdx);
-  }
-
   return spec;
 
 failure:  // on failure free the spec fields array and return an error
@@ -346,10 +402,6 @@ int IndexSpec_AddTerm(IndexSpec *sp, const char *term, size_t len) {
   return isNew;
 }
 
-void IndexSpec_RestoreTerm(IndexSpec *sp, const char *term, size_t len, double score) {
-  Trie_InsertStringBuffer(sp->terms, (char *)term, len, score, 0, NULL);
-}
-
 /// given an array of random weights, return the a weighted random selection, as the index in the
 /// array
 size_t weightedRandom(double weights[], size_t len) {
@@ -371,9 +423,9 @@ size_t weightedRandom(double weights[], size_t len) {
   return 0;
 }
 
-/* Get a random term from the index spec using weighted random. Weighted random is done by sampling
- * N terms from the index and then doing weighted random on them. A sample size of 10-20 should be
- * enough. Returns NULL if the index is empty */
+/* Get a random term from the index spec using weighted random. Weighted random is done by
+ * sampling N terms from the index and then doing weighted random on them. A sample size of 10-20
+ * should be enough. Returns NULL if the index is empty */
 char *IndexSpec_GetRandomTerm(IndexSpec *sp, size_t sampleSize) {
 
   if (sampleSize > sp->terms->size) {
@@ -431,6 +483,20 @@ void IndexSpec_Free(void *ctx) {
     StopWordList_Unref(spec->stopwords);
     spec->stopwords = NULL;
   }
+
+  if (spec->smap) {
+    SynonymMap_Free(spec->smap);
+  }
+
+  if (spec->indexStrs) {
+    for (size_t ii = 0; ii < spec->numFields; ++ii) {
+      if (spec->indexStrs[ii]) {
+        RedisModule_FreeString(spec->strCtx, spec->indexStrs[ii]);
+      }
+    }
+    rm_free(spec->indexStrs);
+    RedisModule_FreeThreadSafeContext(spec->strCtx);
+  }
   rm_free(spec);
 }
 
@@ -460,6 +526,30 @@ IndexSpec *IndexSpec_Load(RedisModuleCtx *ctx, const char *name, int openWrite)
   return IndexSpec_LoadEx(ctx, s, openWrite, NULL);
 }
 
+RedisModuleString *IndexSpec_GetFormattedKey(IndexSpec *sp, const FieldSpec *fs) {
+  if (!sp->indexStrs) {
+    sp->indexStrs = rm_calloc(SPEC_MAX_FIELDS, sizeof(*sp->indexStrs));
+    sp->strCtx = RedisModule_GetThreadSafeContext(NULL);
+  }
+  RedisModuleString *ret = sp->indexStrs[fs->index];
+  if (!ret) {
+    RedisSearchCtx sctx = {.redisCtx = sp->strCtx, .spec = sp};
+    if (fs->type == FIELD_NUMERIC) {
+      ret = fmtRedisNumericIndexKey(&sctx, fs->name);
+    } else if (fs->type == FIELD_TAG) {
+      ret = TagIndex_FormatName(&sctx, fs->name);
+    } else {
+      // Unknown
+      ret = NULL;
+    }
+  }
+  if (!ret) {
+    return NULL;
+  }
+  sp->indexStrs[fs->index] = ret;
+  return ret;
+}
+
 t_fieldMask IndexSpec_ParseFieldMask(IndexSpec *sp, RedisModuleString **argv, int argc) {
   t_fieldMask ret = 0;
 
@@ -473,6 +563,13 @@ t_fieldMask IndexSpec_ParseFieldMask(IndexSpec *sp, RedisModuleString **argv, in
   return ret;
 }
 
+void IndexSpec_InitializeSynonym(IndexSpec *sp) {
+  if (!sp->smap) {
+    sp->smap = SynonymMap_New(false);
+    sp->flags |= Index_HasSmap;
+  }
+}
+
 int IndexSpec_ParseStopWords(IndexSpec *sp, RedisModuleString **strs, size_t len) {
   // if the index already has custom stopwords, let us free them first
   if (sp->stopwords && sp->flags & Index_HasCustomStopwords) {
@@ -500,16 +597,14 @@ int IndexSpec_IsStopWord(IndexSpec *sp, const char *term, size_t len) {
 }
 
 IndexSpec *NewIndexSpec(const char *name, size_t numFields) {
-  IndexSpec *sp = rm_malloc(sizeof(IndexSpec));
+  IndexSpec *sp = rm_calloc(1, sizeof(IndexSpec));
   sp->fields = rm_calloc(sizeof(FieldSpec), numFields ? numFields : SPEC_MAX_FIELDS);
-  sp->numFields = 0;
+  sp->sortables = NewSortingTable();
   sp->flags = INDEX_DEFAULT_FLAGS;
   sp->name = rm_strdup(name);
-  sp->docs = NewDocTable(1000);
+  sp->docs = DocTable_New(100);
   sp->stopwords = DefaultStopWordList();
   sp->terms = NewTrie();
-  sp->sortables = NULL;
-  sp->gc = NULL;
   memset(&sp->stats, 0, sizeof(sp->stats));
   return sp;
 }
@@ -521,7 +616,7 @@ void IndexSpec_StartGC(RedisModuleCtx *ctx, IndexSpec *sp, float initialHZ) {
   if (sp->gc == NULL && RSGlobalConfig.enableGC) {
     RedisModuleString *keyName = RedisModule_CreateString(ctx, sp->name, strlen(sp->name));
     RedisModule_RetainString(ctx, keyName);
-    sp->gc = NewGarbageCollector(keyName, initialHZ);
+    sp->gc = NewGarbageCollector(keyName, initialHZ, sp->unique_id);
     GC_Start(sp->gc);
     RedisModule_Log(ctx, "verbose", "Starting GC for index %s", sp->name);
   }
@@ -538,7 +633,7 @@ int bit(t_fieldMask id) {
 }
 
 // Backwards compat version of load for rdbs with version < 8
-void __fieldSpec_rdbLoadCompat8(RedisModuleIO *rdb, FieldSpec *f, int encver) {
+static void FieldSpec_RdbLoadCompat8(RedisModuleIO *rdb, FieldSpec *f, int encver) {
 
   f->name = RedisModule_LoadStringBuffer(rdb, NULL);
   // the old versions encoded the bit id of the field directly
@@ -559,7 +654,7 @@ void __fieldSpec_rdbLoadCompat8(RedisModuleIO *rdb, FieldSpec *f, int encver) {
   }
 }
 
-void __fieldSpec_rdbSave(RedisModuleIO *rdb, FieldSpec *f) {
+static void FieldSpec_RdbSave(RedisModuleIO *rdb, FieldSpec *f) {
   RedisModule_SaveStringBuffer(rdb, f->name, strlen(f->name) + 1);
   RedisModule_SaveUnsigned(rdb, f->type);
   RedisModule_SaveUnsigned(rdb, f->options);
@@ -574,11 +669,11 @@ void __fieldSpec_rdbSave(RedisModuleIO *rdb, FieldSpec *f) {
   }
 }
 
-void __fieldSpec_rdbLoad(RedisModuleIO *rdb, FieldSpec *f, int encver) {
+static void FieldSpec_RdbLoad(RedisModuleIO *rdb, FieldSpec *f, int encver) {
 
   // Fall back to legacy encoding if needed
   if (encver < INDEX_MIN_TAGFIELD_VERSION) {
-    return __fieldSpec_rdbLoadCompat8(rdb, f, encver);
+    return FieldSpec_RdbLoadCompat8(rdb, f, encver);
   }
 
   f->name = RedisModule_LoadStringBuffer(rdb, NULL);
@@ -604,7 +699,7 @@ void __fieldSpec_rdbLoad(RedisModuleIO *rdb, FieldSpec *f, int encver) {
   }
 }
 
-void __indexStats_rdbLoad(RedisModuleIO *rdb, IndexStats *stats) {
+static void IndexStats_RdbLoad(RedisModuleIO *rdb, IndexStats *stats) {
   stats->numDocuments = RedisModule_LoadUnsigned(rdb);
   stats->numTerms = RedisModule_LoadUnsigned(rdb);
   stats->numRecords = RedisModule_LoadUnsigned(rdb);
@@ -617,7 +712,7 @@ void __indexStats_rdbLoad(RedisModuleIO *rdb, IndexStats *stats) {
   stats->termsSize = RedisModule_LoadUnsigned(rdb);
 }
 
-void __indexStats_rdbSave(RedisModuleIO *rdb, IndexStats *stats) {
+static void IndexStats_RdbSave(RedisModuleIO *rdb, IndexStats *stats) {
   RedisModule_SaveUnsigned(rdb, stats->numDocuments);
   RedisModule_SaveUnsigned(rdb, stats->numTerms);
   RedisModule_SaveUnsigned(rdb, stats->numRecords);
@@ -635,10 +730,10 @@ void *IndexSpec_RdbLoad(RedisModuleIO *rdb, int encver) {
     return NULL;
   }
   RedisModuleCtx *ctx = RedisModule_GetContextFromIO(rdb);
-  IndexSpec *sp = rm_malloc(sizeof(IndexSpec));
+  IndexSpec *sp = rm_calloc(1, sizeof(IndexSpec));
+  sp->sortables = NewSortingTable();
   sp->terms = NULL;
-  sp->docs = NewDocTable(1000);
-  sp->sortables = NULL;
+  sp->docs = DocTable_New(1000);
   sp->name = RedisModule_LoadStringBuffer(rdb, NULL);
   sp->gc = NULL;
   sp->flags = (IndexFlags)RedisModule_LoadUnsigned(rdb);
@@ -650,21 +745,18 @@ void *IndexSpec_RdbLoad(RedisModuleIO *rdb, int encver) {
   sp->fields = rm_calloc(sp->numFields, sizeof(FieldSpec));
   int maxSortIdx = -1;
   for (int i = 0; i < sp->numFields; i++) {
-
-    __fieldSpec_rdbLoad(rdb, &sp->fields[i], encver);
+    FieldSpec *fs = sp->fields + i;
+    FieldSpec_RdbLoad(rdb, sp->fields + i, encver);
     sp->fields[i].index = i;
-
-    /* keep track of sorting indexes to rebuild the table */
-    if (sp->fields[i].sortIdx > maxSortIdx) {
-      maxSortIdx = sp->fields[i].sortIdx;
+    if (FieldSpec_IsSortable(fs)) {
+      assert(fs->sortIdx < RS_SORTABLES_MAX);
+      sp->sortables->fields[fs->sortIdx].name = fs->name;
+      sp->sortables->fields[fs->sortIdx].type = fieldTypeToValueType(fs->type);
+      sp->sortables->len = MAX(sp->sortables->len, fs->sortIdx + 1);
     }
   }
-  /* if we have sortable fields - rebuild the sorting table */
-  if (maxSortIdx >= 0) {
-    _spec_buildSortingTable(sp, maxSortIdx + 1);
-  }
 
-  __indexStats_rdbLoad(rdb, &sp->stats);
+  IndexStats_RdbLoad(rdb, &sp->stats);
 
   DocTable_RdbLoad(&sp->docs, rdb, encver);
   /* For version 3 or up - load the generic trie */
@@ -680,7 +772,20 @@ void *IndexSpec_RdbLoad(RedisModuleIO *rdb, int encver) {
     sp->stopwords = DefaultStopWordList();
   }
 
+  sp->unique_id = spec_unique_ids++;
+
   IndexSpec_StartGC(ctx, sp, GC_DEFAULT_HZ);
+  RedisModuleString *specKey = RedisModule_CreateStringPrintf(ctx, INDEX_SPEC_KEY_FMT, sp->name);
+  CursorList_AddSpec(&RSCursors, sp->name, RSCURSORS_DEFAULT_CAPACITY);
+  RedisModule_FreeString(ctx, specKey);
+
+  sp->smap = NULL;
+  if (sp->flags & Index_HasSmap) {
+    sp->smap = SynonymMap_RdbLoad(rdb, encver);
+  }
+  if (IndexSpec_OnCreate) {
+    IndexSpec_OnCreate(sp);
+  }
   return sp;
 }
 
@@ -693,10 +798,10 @@ void IndexSpec_RdbSave(RedisModuleIO *rdb, void *value) {
 
   RedisModule_SaveUnsigned(rdb, sp->numFields);
   for (int i = 0; i < sp->numFields; i++) {
-    __fieldSpec_rdbSave(rdb, &sp->fields[i]);
+    FieldSpec_RdbSave(rdb, &sp->fields[i]);
   }
 
-  __indexStats_rdbSave(rdb, &sp->stats);
+  IndexStats_RdbSave(rdb, &sp->stats);
   DocTable_RdbSave(&sp->docs, rdb);
   // save trie of terms
   TrieType_GenericSave(rdb, sp->terms, 0);
@@ -705,110 +810,20 @@ void IndexSpec_RdbSave(RedisModuleIO *rdb, void *value) {
   if (sp->flags & Index_HasCustomStopwords) {
     StopWordList_RdbSave(rdb, sp->stopwords);
   }
-}
 
-static void rewriteAofTerms(RedisModuleIO *io, const char *indexName, TrieNode *root) {
-  TrieIterator *iter = TrieNode_Iterate(root, NULL, NULL, NULL);
-  rune *runeStr;
-  t_len runeStrLen;
-  RSPayload *payload;
-  float score;
-
-  while (TrieIterator_Next(iter, &runeStr, &runeStrLen, NULL, &score, NULL)) {
-    size_t bufLen;
-    char *buf = runesToStr(runeStr, runeStrLen, &bufLen);
-
-    char floatBuf[32] = {0};
-    sprintf(floatBuf, "%f", score);
-
-    RedisModule_EmitAOF(io, "FT.TERMADD", "ccc", indexName, buf, floatBuf);
-    free(buf);
+  if (sp->flags & Index_HasSmap) {
+    SynonymMap_RdbSave(rdb, sp->smap);
   }
-
-  TrieIterator_Free(iter);
 }
 
 void IndexSpec_Digest(RedisModuleDigest *digest, void *value) {
 }
 
-#define __vpushStr(v, ctx, str) Vector_Push(v, RedisModule_CreateString(ctx, str, strlen(str)))
-
-void IndexSpec_AofRewrite(RedisModuleIO *aof, RedisModuleString *key, void *value) {
-  IndexSpec *sp = value;
-  Vector *args = NewVector(RedisModuleString *, 4 + 4 * sp->numFields);
-  RedisModuleCtx *ctx = RedisModule_GetContextFromIO(aof);
-
-  // printf("sp->fags:%x\n", sp->flags);
-  // serialize flags
-  if (!(sp->flags & Index_StoreTermOffsets)) {
-    __vpushStr(args, ctx, SPEC_NOOFFSETS_STR);
-  }
-  if (!(sp->flags & Index_StoreByteOffsets)) {
-    __vpushStr(args, ctx, SPEC_NOHL_STR);
-  }
-  if (!(sp->flags & Index_StoreFieldFlags)) {
-    __vpushStr(args, ctx, SPEC_NOFIELDS_STR);
-  }
-
-  // write SCHEMA keyword
-  __vpushStr(args, ctx, SPEC_SCHEMA_STR);
-
-  // serialize schema
-  for (int i = 0; i < sp->numFields; i++) {
-
-    switch (sp->fields[i].type) {
-      case FIELD_FULLTEXT:
-        __vpushStr(args, ctx, sp->fields[i].name);
-        __vpushStr(args, ctx, SPEC_TEXT_STR);
-        if (sp->fields[i].textOpts.weight != 1.0) {
-          __vpushStr(args, ctx, SPEC_WEIGHT_STR);
-          Vector_Push(args,
-                      RedisModule_CreateStringPrintf(ctx, "%f", sp->fields[i].textOpts.weight));
-        }
-        if (FieldSpec_IsNoStem(&sp->fields[i])) {
-          __vpushStr(args, ctx, SPEC_NOSTEM_STR);
-        }
-        break;
-      case FIELD_NUMERIC:
-        __vpushStr(args, ctx, sp->fields[i].name);
-        __vpushStr(args, ctx, NUMERIC_STR);
-        break;
-      case FIELD_GEO:
-        __vpushStr(args, ctx, sp->fields[i].name);
-        __vpushStr(args, ctx, GEO_STR);
-        break;
-      case FIELD_TAG:
-        __vpushStr(args, ctx, sp->fields[i].name);
-        __vpushStr(args, ctx, SPEC_TAG_STR);
-        break;
-      default:
-        RedisModule_Log(ctx, "warning", "Could not save field(%s,%d) with aof)", sp->fields[i].name,
-                        sp->fields[i].type);
-        break;
-    }
-    if (FieldSpec_IsSortable(&sp->fields[i])) {
-      __vpushStr(args, ctx, SPEC_SORTABLE_STR);
-    }
-  }
-
-  size_t offset = strlen(INDEX_SPEC_KEY_PREFIX);
-  const char *indexName = RedisModule_StringPtrLen(key, NULL);
-  indexName += offset;
-
-  RedisModule_EmitAOF(aof, "FT.CREATE", "cv", indexName, (RedisModuleString *)args->data,
-                      Vector_Size(args));
-
-  DocTable_AOFRewrite(&sp->docs, indexName, aof);
-  rewriteAofTerms(aof, indexName, sp->terms->root);
-
-  Vector_Free(args);
-}
-
 int IndexSpec_RegisterType(RedisModuleCtx *ctx) {
   RedisModuleTypeMethods tm = {.version = REDISMODULE_TYPE_METHOD_VERSION,
                                .rdb_load = IndexSpec_RdbLoad,
                                .rdb_save = IndexSpec_RdbSave,
-                               .aof_rewrite = IndexSpec_AofRewrite,
+                               .aof_rewrite = GenericAofRewrite_DisabledHandler,
                                .free = IndexSpec_Free};
 
   IndexSpecType = RedisModule_CreateDataType(ctx, "ft_index0", INDEX_CURRENT_VERSION, &tm);
