@@ -3,10 +3,20 @@
 #ifndef REDISEARCH_MODULE_VERSION
 
 #define REDISEARCH_VERSION_MAJOR 1
-#define REDISEARCH_VERSION_MINOR 0
-#define REDISEARCH_VERSION_PATCH 7
+#define REDISEARCH_VERSION_MINOR 2
+#define REDISEARCH_VERSION_PATCH 0
 
 #define REDISEARCH_MODULE_VERSION \
   (REDISEARCH_VERSION_MAJOR * 10000 + REDISEARCH_VERSION_MINOR * 100 + REDISEARCH_VERSION_PATCH)
 
+#ifdef RS_GIT_VERSION
+static inline const char* RS_GetExtraVersion() {
+  return RS_GIT_VERSION;
+}
+#else
+static inline const char* RS_GetExtraVersion() {
+  return "";
+}
+#endif
+
 #endif
